"""
Computing the feature vectors is the most computationally expensive part of creating these
timelapses. This module is responsible for offloading that computation to the GPU to improve
throughput.
"""



"""
Common operations on iterators, which are the canonical representation of videos, data sources etc
throughout this application.
"""

import collections
import contextlib
import datetime
import itertools
import logging
import threading
from queue import Full, Queue
from threading import Event
from typing import Deque, Iterator, Tuple, TypeVar

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


def first_item_from_iterator(iterator: Iterator[T]) -> Tuple[T, Iterator[T]]:
    """
    Get the first item out of an iterator to surmise some properties of the rest of the items.
    :param iterator: To preview.
    :return: A tuple:
    (the first item in the iterator, the FULL iterator,
    note that the first item is added back onto this output iterator )
    """

    try:
        first_item = next(iterator)
    except StopIteration:
        LOGGER.error("Iterator source was empty, nothing to preview.")
        raise

    return first_item, itertools.chain([first_item], iterator)


def items_per_second(source: Iterator[T], queue_size: int = 60) -> Iterator[T]:
    """
    Prints logging around how often items are being extracted.
    :param source: To forward.
    :param queue_size: Average is computed across this many items.
    :return: The input iterator.
    """

    queue_count = itertools.count()
    item_queue: Deque[datetime.datetime] = collections.deque(maxlen=queue_size)

    def yield_item(item: T) -> T:
        """
        Return the input item, printing speed logging along the way.
        :param item: To forward.
        :return: The input item, unmodified.
        """
        item_queue.append(datetime.datetime.now())
        if next(queue_count) >= queue_size:
            LOGGER.info(
                f"The last {queue_size} items were consumed at a rate of: "
                f"{queue_size / ((item_queue[-1] - item_queue[0]).total_seconds())} items per "
                "second."
            )

        # Don't do anything to the input item.
        return item

    return map(yield_item, source)


def preload_into_memory(
    source: Iterator[T], buffer_size: int, fill_buffer_before_yield: bool = False
) -> Iterator[T]:
    """
    Iterators that involve reading from disk can be slow. To make sure that consumers always
    have a nice chunk of items to consume from RAM, this function creates a worker thread that
    reads items from source and puts them in a Queue. When the output is iterated on, it is
    getting items from that queue, not from the input iterator.
    :param source: Input iterator.
    :param buffer_size: Size of the output queue. If the objects in `source` are big, this is going
    to decide how much RAM is allocated for this process.
    :param fill_buffer_before_yield: Blocks until the intermediate queue has been filled completely
    the first time. This gives consumers a chunk of in-memory frames to work off of before
    having to wait around on IO again.
    :return: An iterator of the buffered items from `source`.
    """

    buffer_filled = Event()
    item_queue: "Queue[T | object]" = Queue(maxsize=buffer_size)
    sentinel = object()  # Unique object to signal the end of the iterator

    def worker() -> None:
        """
        Read items from `source`, putting it into the Queue.
        `.put` operations block if the queue is full, which is used to make sure it is topped up.
        :return: None
        """

        with contextlib.suppress(Exception):  # pylint: disable=broad-except
            for input_item in source:

                if fill_buffer_before_yield and not buffer_filled.is_set():
                    try:
                        item_queue.put_nowait(input_item)
                    except Full:
                        buffer_filled.set()
                else:
                    item_queue.put(input_item)

                del input_item

        # Unblock consumer in the case that the input iterator is smaller than the buffer size.
        buffer_filled.set()

        item_queue.put(sentinel)

    # Start the background thread to read from the input and fill the queue.
    thread = threading.Thread(target=worker, daemon=True)
    thread.start()

    waiting_on_buffer = fill_buffer_before_yield

    # Iterator that consumes from the queue
    while True:

        if waiting_on_buffer:
            LOGGER.debug("Waiting for queue to fill.")
            buffer_filled.wait()
            LOGGER.debug("Queue full, yielding items.")
            waiting_on_buffer = False

        output_item = item_queue.get()
        if output_item is sentinel:
            break
        yield output_item  # type: ignore

    thread.join()



"""Makes looking up asset locations easier in other code"""

from pathlib import Path

_ASSETS_DIRECTORY = Path(__file__).parent.resolve()

SIMPLE_THUMBNAIL_OVERLAY_TEMPLATE = (
    _ASSETS_DIRECTORY / "simple_thumbnail_overlay_template.html.jinja2"
)



"""
Common functionality for dealing with video files.
"""

import errno
import logging
import math
import os
import pprint
import select
import subprocess
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, Iterator, List, NamedTuple, Optional, Tuple, Union, cast

import cv2
import ffmpeg
import more_itertools
import numpy as np
from ffmpeg.nodes import FilterableStream

from content_aware_timelapse.viderator import image_common
from content_aware_timelapse.viderator.iterator_common import first_item_from_iterator
from content_aware_timelapse.viderator.viderator_types import (
    ImageResolution,
    ImageSourceType,
    RectangleRegion,
    RGBInt8ImageType,
)

_DEFAULT_WINDOW_NAME = "Frames"

LOGGER = logging.getLogger(__name__)

logging.getLogger("WriteGear").setLevel(logging.ERROR)


def _write_video_with_audio(video_path: Path, audio: FilterableStream, output_path: Path) -> None:
    """
    Adds an audio file to a video file. Copied from: https://stackoverflow.com/a/65547166
    :param video_path: Path to the video file.
    :param audio: The ffmpeg representation of the audio.
    :param output_path: The path to write the new file to.
    :return: None
    """
    ffmpeg.run(
        ffmpeg.output(
            ffmpeg.input(str(video_path)).video,
            audio,
            str(output_path),
            vcodec="copy",
            acodec="aac",
            audio_bitrate="192k",
            strict="experimental",
            shortest=None,  # <- important!
        ),
        quiet=True,
        overwrite_output=True,
    )


def _read_wav(audio_path: Path) -> FilterableStream:
    """
    Read an audio file as an ffmpeg stream.
    :param audio_path: Path to the audio file.
    :return: The ffmpeg stream of the audio.
    """
    return ffmpeg.input(str(audio_path)).audio


def add_wav_to_video(video_path: Path, audio_path: Path, output_path: Path) -> None:
    """
    Adds an audio file to a video file. Copied from: https://stackoverflow.com/a/65547166
    :param video_path: Path to the video file.
    :param audio_path: Path to the audio file.
    :param output_path: The path to write the new file to.
    :return: None
    """
    _write_video_with_audio(
        video_path=video_path, audio=_read_wav(audio_path), output_path=output_path
    )


def add_wavs_to_video(video_path: Path, audio_paths: List[Path], output_path: Path) -> None:
    """
    Adds an audio file to a video file. Copied from: https://stackoverflow.com/a/65547166
    :param video_path: Path to the video file.
    :param audio_paths: Paths to multiple audio files.
    :param output_path: The path to write the new file to.
    :return: None
    """
    _write_video_with_audio(
        video_path=video_path,
        audio=ffmpeg.concat(*[_read_wav(audio_path) for audio_path in audio_paths], v=0, a=1),
        output_path=output_path,
    )


class VideoOutputController(NamedTuple):
    """
    Interface for something to write videos with.
    Mimics `cv2.VideoWriter`.
    """

    # Adds a frame to the video.
    write: Callable[[RGBInt8ImageType], None]

    # Finalizes the video, you should be able to open and read the video after calling this.
    release: Callable[[], None]


def _create_video_writer_resolution(
    video_path: Path, video_fps: float, resolution: ImageResolution, high_quality: bool
) -> VideoOutputController:
    """
    Create a video writer of a given FPS and resolution.
    :param video_path: Resulting file path.
    :param video_fps: FPS of the video.
    :param resolution: Size of the resulting video.
    :param high_quality: If true, `ffmpeg` will be invoked directly this will create a much larger,
    much higher quality output. TODO: Need to have different writers for the different types of
    outputs.
    :return: The writer.
    """

    if high_quality:

        if resolution.height % 2 != 0 or resolution.width % 2 != 0:
            raise ValueError(
                "Need even video resolution for 420 chroma subsampling. "
                f"Input resolution: {resolution}"
            )

        # Originally used `-movflags +faststart` vs `frag_keyframe+empty_moov` but ran into problems
        # writing to slow filesystems.

        ffmpeg_cmd = [
            "ffmpeg",
            "-y",
            "-loglevel",
            "error",
            "-f",
            "rawvideo",
            "-pix_fmt",
            "rgb24",
            "-s",
            f"{resolution.width}x{resolution.height}",
            "-r",
            str(video_fps),
            "-i",
            "-",
            "-c:v",
            "libx264",
            "-crf",
            "18",
            "-preset",
            "veryslow",
            "-pix_fmt",
            "yuv420p",
            "-movflags",
            "frag_keyframe+empty_moov",
            str(video_path),
        ]

        ffmpeg_proc = subprocess.Popen(  # pylint: disable=consider-using-with
            ffmpeg_cmd,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=0,
        )

        stdin_fd = ffmpeg_proc.stdin.fileno()

        def write_frame(image: RGBInt8ImageType, write_frame_timeout: float = 5.0) -> None:
            """
            Write a frame to the FFmpeg process with a timeout.

            :param image: Frame to write.
            :param write_frame_timeout: Max seconds to wait for the pipe to be ready between frame
            writes.
            :raises TimeoutError: If the pipe is not ready within the timeout.
            :raises RuntimeError: If FFmpeg closes the pipe unexpectedly.
            """
            if image.shape[1] != resolution.width or image.shape[0] != resolution.height:
                raise ValueError("Incoming frame did not match output resolution!")
            if image.dtype != np.uint8:
                raise ValueError("Input image must be uint8.")

            color_space_swapped = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

            # Wait until the pipe is ready for writing, with timeout
            # select.select(read_list, write_list, exception_list, timeout)
            # rlist: file descriptors ready for reading -> empty
            # wlist: file descriptors ready for writing
            # xlist: file descriptors with exceptional conditions -> empty
            _rlist, wlist, _xlist = select.select([], [stdin_fd], [], write_frame_timeout)

            if not wlist:
                # stdin_fd was not ready for writing within timeout
                raise TimeoutError(f"FFmpeg stdin not ready after {write_frame_timeout} seconds")

            try:
                # Write frame data to FFmpeg stdin
                ffmpeg_proc.stdin.write(color_space_swapped.tobytes())
            except OSError as e:
                if e.errno == errno.EPIPE:
                    # FFmpeg process closed stdin unexpectedly
                    raise RuntimeError("FFmpeg process stdin closed unexpectedly") from e
                raise

        def release() -> None:
            """
            Properly finishes the FFmpeg pipeline and ensures the MP4 finalizes.
            :return: None
            """

            if ffmpeg_proc.stdin and not ffmpeg_proc.stdin.closed:
                try:
                    ffmpeg_proc.stdin.flush()
                    ffmpeg_proc.stdin.close()
                except Exception as _exn:  # pylint: disable=broad-except
                    LOGGER.exception("Ran into exception trying to flush/close FFmpeg process.")

            if ffmpeg_proc.stderr:
                try:
                    err = ffmpeg_proc.stderr.read().decode("utf-8", errors="replace")
                    if err:
                        LOGGER.error(f"FFmpeg process produced stderr: {err}")
                except Exception as _exn:  # pylint: disable=broad-except
                    LOGGER.exception(
                        "Ran into exception trying to read stderr from FFmpeg process."
                    )

            return_code = ffmpeg_proc.wait()

            if return_code != 0:
                LOGGER.error(f"FFmpeg exited with code {return_code}")

        return VideoOutputController(write=write_frame, release=release)

    else:
        opencv_writer = cv2.VideoWriter(
            str(video_path),
            cv2.VideoWriter_fourcc(*"mp4v"),  # type: ignore
            video_fps,
            (resolution.width, resolution.height),
        )

        def frame_writer(image: RGBInt8ImageType) -> None:
            """
            Helper function to satisfy mypy.
            Adds a check to make sure the input image matches the expected
            resolution of the output. If this doesn't happen, openCV will
            silently create a video without frames.
            :param image: To write.
            :return: None
            """
            if image_common.image_resolution(image) != resolution:
                raise ValueError("Incoming frame did not match output resolution!")
            opencv_writer.write(image)

        return VideoOutputController(write=frame_writer, release=opencv_writer.release)


def create_video_writer(  # pylint: disable=too-many-positional-arguments
    video_path: Path,
    video_fps: float,
    video_height: int,
    num_squares_width: int,
    num_squares_height: int = 1,
    high_quality: bool = False,
) -> VideoOutputController:
    """
    Helper function to configure the VideoWriter which writes frames to a video file.
    :param video_path: Path to the file on disk.
    :param video_fps: Desired FPS of the video.
    :param video_height: Height of the video in pixels.
    :param num_squares_width: Since each section of the video is a `video_height` x `video_height`
    square this parameter sets the width for the video in pixels, with the number of these squares
    that will be written in each frame.
    :param num_squares_height: Like `num_squares_width`, but for height. Sets the height of
    the video in units of `video_height`.
    :param high_quality: Will be forwarded to library function.
    :return: The openCV `VideoWriter` object.
    """

    return _create_video_writer_resolution(
        video_path=video_path,
        video_fps=video_fps,
        resolution=ImageResolution(
            width=video_height * num_squares_width, height=video_height * num_squares_height
        ),
        high_quality=high_quality,
    )


class VideoFrames(NamedTuple):
    """
    Contains metadata about the video, and an iterator that produces the frames.
    """

    original_fps: float
    total_frame_count: int
    original_resolution: ImageResolution
    frames: ImageSourceType


def divide_no_remainder(numerator: Union[int, float], denominator: Union[int, float]) -> int:
    """
    Raise a Value Error if the division is not even.
    Return `numerator` / `denominator`.
    :param numerator: See top.
    :param denominator: See top.
    :return: `numerator` / `denominator`.
    """

    frac, repeats = math.modf(numerator / denominator)

    if frac != 0:
        raise ValueError(f"Cannot evenly divide {numerator} into {denominator}")

    return int(repeats)


def reduce_fps_take_every(original_fps: float, new_fps: float) -> Optional[int]:
    """
    When reducing a video from a high FPS, to a lower FPS, you should take every nth (output)
    frame of the input video.
    TODO: Wording very rough here.
    :param original_fps: Must be higher than `new_fps`.
    :param new_fps: Must go evenly into `original_fps`.
    :return: Take every n frames to get an evenly reduced output video.
    """

    if new_fps is not None:

        whole = divide_no_remainder(numerator=original_fps, denominator=new_fps)

        if whole != 1:
            return int(whole)

    return None


def write_source_to_disk_forward(
    source: ImageSourceType,
    video_path: Path,
    video_fps: float,
    audio_paths: Optional[List[Path]] = None,
    high_quality: bool = False,
) -> ImageSourceType:
    """
    Consume an image source, write it out to disk.
    :param source: To write to disk.
    :param video_path: Output video path.
    :param video_fps: Frames/Second of the output video.
    :param audio_paths: If given, the audio files will be written to the output video.
    :param high_quality: Flag will be forwarded to library function.
    :return: None
    """

    def setup_iteration(output_path: Path) -> ImageSourceType:
        """
        Helper function to set up the output and forwarding operation.
        :param output_path: Intermediate video path.
        :return: The frames to yield.
        """

        first_frame, frame_source = first_item_from_iterator(source)

        writer = _create_video_writer_resolution(
            video_path=output_path,
            video_fps=video_fps,
            resolution=image_common.image_resolution(first_frame),
            high_quality=high_quality,
        )

        def write_frame(frame: RGBInt8ImageType) -> None:
            """
            Write the given frame to the file.
            :param frame: To write.
            :return: None
            """
            rgb_frame = cast(
                RGBInt8ImageType, cv2.cvtColor(frame.astype(np.uint8), cv2.COLOR_BGR2RGB)
            )
            writer.write(rgb_frame)

        try:
            for index, image in enumerate(frame_source):
                LOGGER.debug(f"Writing frame #{index} to file: {video_path}")
                write_frame(image)
                yield image
        finally:
            writer.release()

    try:
        if audio_paths is None or not audio_paths:
            yield from setup_iteration(video_path)
        else:
            with video_safe_temp_path(suffix=video_path.suffix) as temp_video_path:
                yield from setup_iteration(temp_video_path)
                LOGGER.info(f"Finalizing {video_path}")
                add_wavs_to_video(
                    video_path=temp_video_path,
                    audio_paths=audio_paths,
                    output_path=video_path,
                )
    except Exception as e:
        LOGGER.exception(f"Ran into an exception writing out a video: {pprint.pformat(e)}")
        raise e


def write_source_to_disk_consume(
    source: ImageSourceType,
    video_path: Path,
    video_fps: float,
    audio_paths: Optional[List[Path]] = None,
    high_quality: bool = False,
) -> None:
    """
    Consume an image source, write it out to disk.
    :param source: To write to disk.
    :param video_path: Output video path.
    :param video_fps: FPS of the output video.
    :param audio_paths: If given, the audio file at this path will be written to the output video.
    :param high_quality: Flag will be forwarded to library function.
    :return: None
    """

    more_itertools.consume(
        write_source_to_disk_forward(
            source=source,
            video_path=video_path,
            video_fps=video_fps,
            audio_paths=audio_paths,
            high_quality=high_quality,
        )
    )


def resize_source(source: ImageSourceType, resolution: ImageResolution) -> ImageSourceType:
    """
    For each item in the input source, scale it to the given resolution.
    :param source: Contains Images.
    :param resolution: Desired resolution.
    :return: A new source of scaled images.
    """

    yield from (
        (
            image_common.resize_image(image, resolution, delete=True)
            if image_common.image_resolution(image) != resolution
            else image
        )
        for image in source
    )


def scale_square_source_duplicate(
    source: ImageSourceType, output_side_length: int, frame_multiplier: int = 1
) -> ImageSourceType:
    """
    Scale the resolution and number of frames in a given source.
    :param source: To scale.
    :param output_side_length: Square frames will be resized to this side length.
    :param frame_multiplier: Every frame will be duplicated this many times.
    :return: Scaled source.
    """

    resized = resize_source(source, ImageResolution(output_side_length, output_side_length))

    return (
        cast(
            ImageSourceType,
            more_itertools.repeat_each(
                resized,
                frame_multiplier,
            ),
        )
        if frame_multiplier != 1
        else resized
    )


def display_frame_forward_opencv(
    source: ImageSourceType,
    window_name: str = _DEFAULT_WINDOW_NAME,
    display_resolution: Optional[ImageResolution] = None,
    full_screen: bool = False,
) -> ImageSourceType:
    """
    Displays the images in `source`, and forwards the image so they can be consumed again.
    Uses an openCV `imshow` to reveal the image.
    :param source: To display.
    :param window_name: Name of the window.
    :param display_resolution: Change the input frames to this resolution before displaying. Doesn't
    modify the output frames.
    :param full_screen: If True, images will be displayed fullscreen, windowed if otherwise.
    :return: Forwarded iterator, `source`.
    """

    cv2.namedWindow(
        window_name,
        cv2.WINDOW_GUI_NORMAL if full_screen else cv2.WINDOW_AUTOSIZE,
    )

    if full_screen:
        cv2.setWindowProperty(window_name, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)

    def display_frame(frame: RGBInt8ImageType) -> RGBInt8ImageType:
        """
        Helper function, actually show the image.
        :param frame: To display.
        :return: `frame`.
        """
        cv2.imshow(
            window_name,
            cv2.cvtColor(
                (
                    image_common.resize_image(frame, display_resolution)
                    if display_resolution is not None
                    else frame
                ),
                cv2.COLOR_RGB2BGR,
            ),
        )
        cv2.waitKey(1)
        return frame

    # yield from is used, so we can run things after the input has been exhausted to cleanup.
    yield from map(display_frame, source)

    cv2.destroyWindow(window_name)


def crop_source(source: ImageSourceType, region: RectangleRegion) -> ImageSourceType:
    """
    Crop all frames in a source to a given region.
    :param source: To crop.
    :param region: Region to crop to, this is an absolute region in the input frame, not a general
    resolution.
    :return: Cropped frames.
    """

    yield from (
        (image_common.crop_image(image=image, region=region, delete=True)) for image in source
    )


@contextmanager
def video_safe_temp_path(suffix: str = ".mp4") -> Iterator[Path]:
    """
    Context manager that yields a Path to a temporary file.
    The file is created and closed immediately, then unlinked
    automatically on context exit.

    :param suffix: Optional file suffix (e.g. '.mp4')
    """
    fd, temp_name = tempfile.mkstemp(suffix=suffix)
    os.close(fd)  # close the low-level fd, we only care about the path
    video_path = Path(temp_name)
    try:
        yield video_path
    finally:
        try:
            video_path.unlink(missing_ok=True)
        except Exception:  # pylint: disable=broad-except
            pass


def concat_videos_for_youtube(video_paths: Tuple[Path, ...], output_path: Path) -> None:
    """
    Concatenates or re-encodes videos with optimal settings for YouTube.
    If a single input is provided, it simply re-encodes that video.

    :param video_paths: Videos to concatenate as a tuple.
    :param output_path: Destination path.
    :raises ValueError: If any input file is missing, duplicated, or mismatched in resolution/FPS.
    """

    if not video_paths:
        raise ValueError("No input files provided.")

    resolved_paths = [p.absolute() for p in video_paths]

    # Validation
    for p in resolved_paths:
        if not p.exists():
            raise ValueError(f"Input file {p} does not exist.")

    if len(resolved_paths) != len(set(resolved_paths)):
        raise ValueError("Duplicate input files are not allowed.")

    widths, heights, fps_set = set(), set(), set()
    for f in resolved_paths:
        stream = ffmpeg.probe(str(f), select_streams="v")["streams"][0]
        widths.add(int(stream["width"]))
        heights.add(int(stream["height"]))
        a, b = stream["r_frame_rate"].split("/")
        fps_set.add(float(int(a) / int(b)))

    if len(widths) > 1 or len(heights) > 1 or len(fps_set) > 1:
        raise ValueError(
            "Input files need to have the same resolution and fps. "
            f"Widths: {widths}, Heights: {heights}, FPS: {fps_set}"
        )

    # Single file (re-encode only)
    if len(resolved_paths) == 1:
        (
            ffmpeg.input(str(resolved_paths[0]))
            .output(
                str(output_path),
                vcodec="libx264",
                crf=18,
                preset="slow",
                pix_fmt="yuv420p",
                acodec="aac",
                audio_bitrate="192k",
                movflags="+faststart",
            )
            .run(overwrite_output=True, quiet=True)
        )
        return

    # Multiple files (concat demuxer)
    with tempfile.NamedTemporaryFile("w", delete=True, suffix=".txt") as list_file:
        for f in resolved_paths:
            list_file.write(f"file '{f}'\n")
        list_file.flush()

        (
            ffmpeg.input(str(list_file.name), format="concat", safe=0)
            .output(
                str(output_path),
                vcodec="libx264",
                crf=18,
                preset="slow",
                pix_fmt="yuv420p",
                acodec="aac",
                audio_bitrate="192k",
                movflags="+faststart",
            )
            .run(overwrite_output=True, quiet=True)
        )


def drop_audio_from_video(input_video: Path, output_video: Path) -> None:
    """
    Removes the audio track from a video file while preserving video quality.

    :param input_video: Path to the input video file.
    :param output_video: Path to save the output video file (no audio).
    :raises ValueError: If the input file does not exist.
    """

    if not input_video.exists():
        raise ValueError(f"Input file {input_video} does not exist.")

    (
        ffmpeg.input(str(input_video))
        .output(
            str(output_video),
            c="copy",  # copy video stream directly (no re-encode)
            an=None,  # drop all audio streams
        )
        .run(overwrite_output=True, quiet=True)
    )



"""
Common functionality and types used in still images and in video.
"""

import logging
import math
from numbers import Integral
from pathlib import Path
from typing import List, Tuple, cast

import cv2
import numpy as np
from PIL import Image, ImageDraw

from content_aware_timelapse.viderator.viderator_types import (
    AspectRatio,
    ImageResolution,
    RectangleRegion,
    RGBInt8ImageType,
    XYPoint,
)

LOGGER = logging.getLogger(__name__)


def image_resolution(image: RGBInt8ImageType) -> ImageResolution:
    """
    Get an image's resolution.
    :param image: To size.
    :return: Image resolution as an NT.
    """

    return ImageResolution(height=image.shape[0], width=image.shape[1])


def resolution_aspect_ratio(resolution: ImageResolution) -> AspectRatio:
    """
    Compute the aspect ratio of an image's resolution

    :param resolution:
    :return: AspectRatio NT with width and height as floats.
    """

    if isinstance(resolution.width, Integral) and isinstance(resolution.height, Integral):
        g = math.gcd(resolution.width, resolution.height)
        return AspectRatio(width=float(resolution.width // g), height=float(resolution.height // g))
    else:
        return AspectRatio(width=float(resolution.width), height=float(resolution.height))


def region_to_resolution(region: RectangleRegion) -> ImageResolution:
    """
    Converts a RectangleRegion into an ImageResolution.

    :param region: RectangleRegion defining the area.
    :return: ImageResolution with width and height of the region.
    """
    width = region.right - region.left
    height = region.bottom - region.top

    if width <= 0 or height <= 0:
        raise ValueError(f"Invalid RectangleRegion {region}: width and height must be positive.")

    return ImageResolution(width=width, height=height)


def largest_fitting_region(
    source_resolution: ImageResolution,
    aspect_ratio: AspectRatio,
    even_dimensions: bool = False,
) -> ImageResolution:
    """
    Determine the largest region of `source` that fits the input aspect ratio.
    :param source_resolution: Bounds.
    :param aspect_ratio: Desired aspect ratio.
    :param even_dimensions: If True, the output resolution will be rounded down such that the
    dimensions are both even numbers. Useful for encoding.
    :return: Largest resolution.
    """

    # Try width-limited fit
    scale_w = source_resolution.width / aspect_ratio.width
    height_if_full_width = aspect_ratio.height * scale_w

    if height_if_full_width <= source_resolution.height:
        width = source_resolution.width
        height = int(height_if_full_width)
    else:
        # Otherwise height-limited fit
        scale_h = source_resolution.height / aspect_ratio.height
        width = int(aspect_ratio.width * scale_h)
        height = source_resolution.height

    if even_dimensions:
        width -= width % 2
        height -= height % 2

    return ImageResolution(width, height)


def load_rgb_image(path: Path) -> RGBInt8ImageType:
    """
    Loads an image from a path and returns it as an RGB uint8 numpy array.

    :param path: Path to the image file.
    :return: RGB image as a (H, W, 3) uint8 ndarray.
    """
    with Image.open(path) as img:
        img = img.convert("RGB")
        arr = np.asarray(img, dtype=np.uint8)
    return RGBInt8ImageType(arr)


def save_rgb_image(path: Path, image: RGBInt8ImageType) -> None:
    """
    Saves an RGB uint8 numpy array to a file at the given path.

    :param path: Destination file path.
    :param image: RGB image as a (H, W, 3) uint8 ndarray.
    """
    img = Image.fromarray(np.asarray(image, dtype=np.uint8), mode="RGB")
    img.save(path)


def resize_image(
    image: RGBInt8ImageType, resolution: ImageResolution, delete: bool = False
) -> RGBInt8ImageType:
    """
    Resizes an image to the input resolution.
    Uses, `cv2.INTER_CUBIC`, which is visually good-looking but somewhat slow.
    May want to be able to pass this in.
    :param image: To scale.
    :param resolution: Output resolution.
    :param delete: If true, `del` will be used on `image` to force it's memory to be released.
    :return: Scaled image.
    """

    output = cast(
        RGBInt8ImageType,
        cv2.resize(image, (resolution.width, resolution.height), interpolation=cv2.INTER_CUBIC),
    )

    if delete:
        # The image has now been 'consumed', and can't be used again.
        # We delete this frame here to avoid memory leaks.
        # Not really sure if this is needed, but it shouldn't cause harm.
        del image

    # The scaled image.
    return output


def resize_image_max_side(
    image: RGBInt8ImageType, max_side_length: int, delete: bool = False
) -> RGBInt8ImageType:
    """
    Resizes an image such that its largest side is `max_side_length`, preserving aspect ratio.
    Uses `cv2.INTER_LINEAR` for speed.

    :param image: Input image to scale.
    :param max_side_length: Maximum length of the largest side after scaling.
    :param delete: If true, deletes the input image to free memory.
    :return: Scaled image.
    """
    resolution = image_resolution(image)

    # Determine scaling factor
    scale = max_side_length / max(resolution.height, resolution.width)

    # If image is already smaller than max_side_length, don't scale up
    if scale >= 1.0:
        output = image.copy()
    else:
        new_width = int(resolution.width * scale)
        new_height = int(resolution.height * scale)

        try:
            output = cast(
                RGBInt8ImageType,
                cv2.resize(image, (new_width, new_height), interpolation=cv2.INTER_LINEAR),
            )
        except cv2.error as e:
            raise ValueError(
                f"Couldn't resize image. Shape: {image.shape}, "
                f"New Size {(new_width, new_height)}, "
                f"Max Side Length: {max_side_length}, "
                f"Scale: {scale}"
            ) from e

    if delete:
        del image

    return output


def crop_image(
    image: RGBInt8ImageType, region: RectangleRegion, delete: bool = False
) -> RGBInt8ImageType:
    """
    Crops an image to the given SquareRegion.

    :param image: Input image (H, W, 3) with dtype uint8.
    :param region: SquareRegion specifying (top, left, bottom, right).
    :param delete: If true, `del` will be used on `image` to free memory.
    :return: Cropped image as RGBInt8ImageType.
    """
    cropped = cast(
        RGBInt8ImageType,
        image[region.top : region.bottom, region.left : region.right],
    )

    if delete:
        del image

    return cropped


def draw_points_on_image(
    points: List[XYPoint],
    image: RGBInt8ImageType,
    radius: int = 5,
    color: Tuple[int, int, int] = (255, 0, 0),
) -> RGBInt8ImageType:
    """
    Draws a list of points onto the input image for visualization.
    :param points: To draw.
    :param image: Image to draw on.
    :param radius: Radius of point in pixels.
    :param color: Color of point (R, G, B)
    :return: Modified image.
    """

    pil_image = Image.fromarray(image)
    draw = ImageDraw.Draw(pil_image)

    for point in points:

        bbox = [point.x - radius, point.y - radius, point.x + radius, point.y + radius]
        draw.ellipse(bbox, fill=color, outline=color)

    # Always return NumPy array
    return RGBInt8ImageType(np.array(pil_image))


def draw_regions_on_image(
    regions: List[RectangleRegion],
    image: RGBInt8ImageType,
    color: Tuple[int, int, int] = (0, 255, 0),
    width: int = 3,
) -> RGBInt8ImageType:
    """
    Draws rectangular regions onto the input image for visualization.

    :param regions: List of SquareRegion objects to draw.
    :param image: Image to draw on (NumPy array).
    :param color: Color of rectangle (R, G, B)
    :param width: Line width of the rectangle.
    :return: Modified image (NumPy array).
    """
    pil_image = Image.fromarray(image)
    draw = ImageDraw.Draw(pil_image)

    for region in regions:
        bbox = [region.left, region.top, region.right, region.bottom]
        draw.rectangle(bbox, outline=color, width=width)

    return RGBInt8ImageType(np.array(pil_image))


def composite_region(
    destination_image: RGBInt8ImageType,
    destination_region: RectangleRegion,
    image_to_add: RGBInt8ImageType,
) -> RGBInt8ImageType:
    """
    Pastes an image into the destination image at the coordinates defined
    by `destination_region`.

    :param destination_image: The pre-allocated composited image.
    :param destination_region: RectangleRegion defining the top-left corner for placement.
    :param image_to_add: The region to paste into the destination image.
    :return: Updated destination image.
    """

    res = region_to_resolution(destination_region)

    # Ensure the source image matches the destination region size
    if image_to_add.shape[0] != res.height or image_to_add.shape[1] != res.width:
        raise ValueError(
            f"Source image shape {image_to_add.shape[:2]} "
            f"does not match expected region size {(res.width, res.height)}"
        )

    # Paste the source image into the destination image
    destination_image[
        destination_region.top : destination_region.bottom,
        destination_region.left : destination_region.right,
    ] = image_to_add

    return destination_image


def reshape_from_regions(  # pylint: disable=too-many-locals
    image: RGBInt8ImageType,
    prioritized_poi_regions: Tuple[RectangleRegion, ...],
    layout_matrix: List[List[int]],
) -> RGBInt8ImageType:
    """
    Takes region of `image` and composites them in the layout defined by `layout_matrix`, forming
    a new image.

    :param image: The source image to crop from.
    :param prioritized_poi_regions: Flat list of rectangular regions to crop.
    :param layout_matrix: 2D grid of integers representing priorities.
    :return: Composited output image.
    """
    num_rows = len(layout_matrix)
    num_cols = len(layout_matrix[0])

    if len(prioritized_poi_regions) != num_rows * num_cols:
        raise ValueError(
            f"Expected {num_rows * num_cols} regions for a {num_rows}x{num_cols} matrix, "
            f"but got {len(prioritized_poi_regions)}."
        )

    # Determine size of each region
    region_resolution = region_to_resolution(prioritized_poi_regions[0])
    output_width = num_cols * region_resolution.width
    output_height = num_rows * region_resolution.height

    # Create empty canvas
    composited: RGBInt8ImageType = cast(
        RGBInt8ImageType, np.zeros((output_height, output_width, 3), dtype=np.uint8)
    )

    # Reshape flat regions into rows
    row_regions = [
        prioritized_poi_regions[r * num_cols : (r + 1) * num_cols] for r in range(num_rows)
    ]

    # Pair each region with its priority and record its grid position
    regions_with_priority: list[tuple[RectangleRegion, int, int, int]] = []
    for row_idx, (row_region_list, row_priority_list) in enumerate(zip(row_regions, layout_matrix)):
        for col_idx, (region, priority) in enumerate(zip(row_region_list, row_priority_list)):
            regions_with_priority.append((region, priority, row_idx, col_idx))

    # Sort descending by priority
    regions_with_priority.sort(key=lambda rp: rp[1], reverse=True)

    # Composite regions in priority order
    for region, _priority, row_idx, col_idx in regions_with_priority:
        cropped = crop_image(image=image, region=region)

        # Compute destination coordinates in the output canvas
        destination_region = RectangleRegion(
            top=row_idx * region_resolution.height,
            left=col_idx * region_resolution.width,
            bottom=(row_idx + 1) * region_resolution.height,
            right=(col_idx + 1) * region_resolution.width,
        )

        composite_region(
            destination_image=composited,
            destination_region=destination_region,
            image_to_add=cropped,
        )

    return composited



"""
Tries to do `itertools.tee`, but on disk instead of in memory.

Thank u: https://stackoverflow.com/a/70917416
"""

import pickle
import shutil
import threading
from functools import partial
from pathlib import Path
from queue import Queue
from tempfile import NamedTemporaryFile
from typing import Any, Iterator, List, NamedTuple, Optional, Tuple, TypeVar, cast
from typing_extensions import Protocol

import h5py
import hdf5plugin
import more_itertools
import numpy as np
from sentinels import NOTHING

from content_aware_timelapse.viderator import frames_in_video, video_common
from content_aware_timelapse.viderator.viderator_types import ImageSourceType, RGBInt8ImageType

T = TypeVar("T")


class SerializeItem(Protocol):
    """
    Describes a function that writes a given item out to disk.
    """

    def __call__(self, path: Path, items: List[Any]) -> None:  # type: ignore[explicit-any]
        """
        :param path: Path to write the serialized object to on disk.
        :param items: List of objects to serialize.
        :return: None
        """


class DeSerializeItem(Protocol):
    """
    Describes a function that loads an item from disk back into memory.
    """

    def __call__(self, path: Path) -> List[Any]:  # type: ignore[explicit-any]
        """
        :param path: Path to the object on disk.
        :return: Items loaded back into memory.
        """


class Serializer(NamedTuple):
    """
    A pair of functions, one to write and one to load items from disk.
    """

    serialize: SerializeItem
    deserialize: DeSerializeItem


def serialize_pickle(path: Path, items: List[T]) -> None:
    """
    Writes an item to disk using the built-in pickle module.
    :param path: Path to write the serialized object to on disk.
    :param items: List of objects to serialize.
    :return: None
    """

    with open(str(path), "wb") as p:
        pickle.dump(items, p)


def deserialize_pickle(path: Path) -> List[Any]:  # type: ignore[explicit-any]
    """
    Loads a pickled item from disk using the built-in pickle module.
    :param path: Path to the object on disk.
    :return: Item loaded back into memory.
    """

    with open(str(path), "rb") as p:
        return cast(List[Any], pickle.load(p))  # type: ignore[explicit-any]


PICKLE_SERIALIZER = Serializer(serialize=serialize_pickle, deserialize=deserialize_pickle)

HDF5_DATASET_NAME = "item_dataset"


def serialize_hdf5(
    path: Path,
    items: List[RGBInt8ImageType],
    compression: Optional[str],
) -> None:
    """
    Writes an item to disk using hdf5, a format for storing data arrays on disk.
    :param path: Path to write the serialized object to on disk.
    :param items: List of objects to serialize.
    :param compression: Passed to underlying `.create_dataset`
    :param compression_opts: Passed to underlying `.create_dataset`
    :return: None
    """

    items_array = np.stack(items, axis=0)
    with h5py.File(name=str(path), mode="w") as f:
        f.create_dataset(
            HDF5_DATASET_NAME,
            shape=items_array.shape,
            dtype=items_array.dtype,
            data=items_array,
            compression=compression,
        )


def deserialize_hdf5(path: Path) -> List[RGBInt8ImageType]:
    """
    Loads an item to disk using hdf5, a format for storing data arrays on disk.
    :param path: Path to the object on disk.
    :return: Items loaded back into memory.
    """

    with h5py.File(name=str(path), mode="r", libver="latest") as f:
        return [RGBInt8ImageType(frame) for frame in np.array(f[HDF5_DATASET_NAME])]


HDF5_SERIALIZER = Serializer(
    serialize=partial(serialize_hdf5, compression=None),
    deserialize=deserialize_hdf5,
)

HDF5_COMPRESSED_SERIALIZER = Serializer(
    serialize=partial(
        serialize_hdf5,
        compression=hdf5plugin.Blosc2(cname="zstd", clevel=9, filters=hdf5plugin.Blosc2.SHUFFLE),
    ),
    deserialize=deserialize_hdf5,
)


def load_queue_items(queue: "Queue[Path]", deserialize: DeSerializeItem) -> Iterator[T]:
    """
    Iterate over the items in a queue.
    Load the objects on disk back into memory and yield them.
    Before yielding the objects, deletes their source file.
    :param queue: To consume.
    :param deserialize: Function to load the items from disk.
    :return: An iterator of the items stored in the queue.
    """

    for path in iter(queue.get, NOTHING):
        output: List[T] = deserialize(path)
        path.unlink()
        yield from output


def tee_disk_cache(
    iterator: Iterator[T],
    copies: int,
    serializer: Serializer = PICKLE_SERIALIZER,
) -> Tuple[Iterator[T], ...]:
    """
    Caches the results from an input iterator onto disk rather than into memory for re-iteration
    later. Kind of like `itertools.tee`, but instead of going into memory with the copies, the
    intermediate objects are stored on disk.
    :param iterator: The iterator to duplicate.
    :param copies: The number of secondary iterators to make. Think of this like the `n` argument
    to `itertools.tee`.
    :param serializer: Defines how the objects will be stored on disk.
    :return: A tuple:
        (
            The primary iterator. Consume this one to populate the values in the secondary
            iterators.,
            The secondary iterators. When one of these is incremented, its next object
            is loaded from disk and yielded. Note that if you iterate on these past the head of
            `primary`, then the iteration will block.
        )
    """

    path_queues: List["Queue[Path]"] = [Queue() for _ in range(copies)]

    def forward_iterator() -> Iterator[T]:
        """
        Works through the input iterator, and as new times are produced, saves
        them to disk, and fills the queues with their locations.
        :return: Yields the original items from the input iterator.
        """

        for item in iterator:

            # These will get deleted after being loaded into memory later.
            with NamedTemporaryFile(mode="wb", delete=True) as primary_dump:

                primary_path = Path(primary_dump.name)
                serializer.serialize(path=primary_path, items=[item])

                for queue in path_queues:
                    with NamedTemporaryFile(mode="wb", delete=False) as secondary_dump:
                        secondary_path = Path(secondary_dump.name)
                        shutil.copy(src=primary_path, dst=secondary_path)
                        queue.put(secondary_path)

            yield item

        # Tells the queues that no more items will be coming out.
        for queue in path_queues:
            queue.put(NOTHING)

    return (forward_iterator(),) + tuple(
        load_queue_items(queue, deserialize=serializer.deserialize) for queue in path_queues
    )


def disk_buffer(
    source: Iterator[T],
    buffer_size: int,
    serializer: Serializer = HDF5_SERIALIZER,
) -> Iterator[T]:
    """
    Uses a worker thread to read from the source iterator and write the contents to disk. When
    output is required, items are deserialized and written out. Should be used to maintain pipeline
    if input iterator is slow and items are large, disk is used rather than memory to cache the
    items.

    :param source: To buffer and forward.
    :param buffer_size: The target size of the buffer of items on disk.
    :param serializer: Defines how the objects will be stored on disk.
    :return: `source` but having been buffered.
    """

    path_queue: "Queue[Path | object]" = Queue(maxsize=buffer_size)
    sentinel = object()  # Unique object to signal the end of the iterator

    def worker() -> None:
        """
        Read items from the input and serialize them to disk, putting the locations in the queue.
        :return: None
        """

        for chunk_of_items in more_itertools.chunked(source, 250):

            # Don't delete here, we'll delete after consumption.
            with NamedTemporaryFile(mode="wb", delete=False) as dump:
                dump_path = Path(dump.name)
                serializer.serialize(path=dump_path, items=chunk_of_items)
                path_queue.put(dump_path)

        # input has been exhausted.
        path_queue.put(sentinel)

    # Start the background thread to read from the input and fill the queue.
    thread = threading.Thread(target=worker, daemon=True)
    thread.start()

    while True:
        serialized_path = path_queue.get()
        if serialized_path is sentinel:
            break
        yield from serializer.deserialize(path=serialized_path)  # type: ignore

        serialized_path.unlink()  # type: ignore[attr-defined]

    thread.join()


def video_file_tee(
    source: ImageSourceType, copies: int, video_fps: float, intermediate_video_path: Optional[Path]
) -> Tuple[ImageSourceType, ...]:
    """
    Writes the source to disk as a propper video file, then returns iterators reading from that
    file.
    The idea being that video files are space-effecient ways to store video on disk.
    :param source: To copy.
    :param copies: Number of iterators to return.
    :param video_fps: Consumed when writing the intermediate video path.
    :param intermediate_video_path: Path to write the intermediate to.
    :return: A tuple, of length `copies` of iterators, each containing the frames in `source`.
    """

    video_common.write_source_to_disk_consume(
        source=source,
        video_path=intermediate_video_path,
        video_fps=video_fps,
        high_quality=True,
    )

    return tuple(
        frames_in_video.frames_in_video_opencv(video_path=intermediate_video_path).frames
        for _ in range(copies)
    )



"""HTML template rendering helpers."""

# pylint: disable=too-many-positional-arguments

from pathlib import Path
from typing import NewType, Union, cast

from jinja2 import Template

from content_aware_timelapse.viderator.html_on_image import assets

FilledTemplate = NewType("FilledTemplate", str)
"""
NewType here is used to make sure functions that consume filled templates must get these templates
from valid producing functions.
"""


def render_html_template(template_path: Path, **context: Union[str, int, float]) -> FilledTemplate:
    """
    Load and render a Jinja2 HTML document template.

    :param template_path: Path to the template file.
    :param context: Variables to substitute into the template.
    :return: Rendered template string.
    """

    with open(template_path, encoding="utf-8") as fp:
        template = Template(fp.read())

    return cast(FilledTemplate, template.render(context))


def simple_thumbnail_overlay(
    width: int,
    height: int,
    upper_subtitle: str,
    main_title: str,
    lower_title: str,
    gradient_start: str,
    gradient_stop: str,
    text_color: str,
    shadow_color: str,
) -> FilledTemplate:
    """Render the 'simple thumbnail overlay' template."""
    return render_html_template(
        assets.SIMPLE_THUMBNAIL_OVERLAY_TEMPLATE,
        width=width,
        height=height,
        gradient_start=gradient_start,
        gradient_stop=gradient_stop,
        main_title=main_title,
        upper_subtitle=upper_subtitle,
        lower_title=lower_title,
        text_color=text_color,
        shadow_color=shadow_color,
    )



"""
Turn vectorized images to numerical scores, then pick the best images for the output.
"""

import itertools
import logging
from pathlib import Path
from typing import Iterator, List, NamedTuple, Optional, Set, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy import typing as npt
from sklearn import preprocessing

from content_aware_timelapse.frames_to_vectors import conversion
from content_aware_timelapse.frames_to_vectors.conversion import IntermediateFileInfo
from content_aware_timelapse.frames_to_vectors.conversion_types import (
    ConversionScoringFunctions,
    IndexScores,
    ScoreWeights,
)
from content_aware_timelapse.gpu_discovery import GPUDescription
from content_aware_timelapse.viderator import image_common, video_common
from content_aware_timelapse.viderator.viderator_types import ImageSourceType, RGBInt8ImageType

LOGGER = logging.getLogger(__name__)


def _apply_radius_deselection(
    indices_sorted_by_score: List[int],
    num_output_frames: int,
    initial_radius: int,
    minimum_radius: int = 0,
) -> List[int]:
    """
    Tries to remove frames nearby to frames with high scores to avoid clustering.

    :param indices_sorted_by_score: Vector (frame) indices sorted by score, with the most
    interesting frames first.
    :param num_output_frames: The desired number of output frames requested by the user.
    :param initial_radius: The number of frames before/after high scoring frames that should be
    dropped. The function will half this value until `num_output_frames` is reached.
    :param minimum_radius: The smallest value the `initial_radius` can shink to.
    :return: A list of frame indices.
    """

    all_indices = np.array(sorted(indices_sorted_by_score))

    def try_deselection(selection_radius: int) -> List[int]:
        """
        Runs the selection process given the current radius value.
        :param selection_radius: The number of frames to drop before/after high value frames.
        :return: A list of suitable indices.
        """

        suppression_mask = np.full(len(all_indices), True)
        local_selection = []

        for idx in indices_sorted_by_score:

            if not suppression_mask[all_indices == idx][0]:
                continue

            local_selection.append(idx)

            suppression_mask &= np.abs(all_indices - idx) > selection_radius

            if len(local_selection) == num_output_frames:
                break

        return local_selection

    # Try decreasing radii until enough frames are selected
    radius = initial_radius  # Initial suppression radius
    selected_indices = []

    while radius >= minimum_radius:
        selected_indices = try_deselection(radius)

        if len(selected_indices) >= num_output_frames:
            break

        radius = max(radius // 2, minimum_radius)

    return selected_indices[:num_output_frames]  # Just in case we over-selected


class ScoredFrames(NamedTuple):
    """
    Intermediate type.
    """

    best_frame_index: int
    winning_indices: List[int]


def _score_and_sort_frames(
    score_weights: ScoreWeights,
    index_scores: List[IndexScores],
    num_output_frames: int,
    deselection_radius_frames: int,
    plot_path: Optional[Path],
) -> ScoredFrames:
    """
    Selects the top `num_output_frames` while avoiding clusters.

    :param score_weights: Weights applied to the different properties after they are normalized from
    0-1. The overall score is the sum of the weighted properties.
    :param index_scores: List of dictionaries containing frame index and feature scores.
    :param num_output_frames: Number of frames to select.
    :param deselection_radius_frames: Frames surrounding high scoring frames removed to
    prevent clustering. This is the number of frames before/after a high scoring one that are
    slightly decreased in score.
    :param plot_path: If given, creates an overall visualization of the math that went into
    selecting the frames.
    :return: List of selected frame indices.
    """

    def _basic_score(index_score: IndexScores) -> float:
        """
        Computes an overall score for a row given the weighted components of the underlying score.
        :param index_score: From the row, the input parts of the score.
        :return: The overall score.
        """

        return (
            (index_score["entropy"] * score_weights.low_entropy)
            + (index_score["variance"] * score_weights.variance)
            + (index_score["saliency"] * score_weights.saliency)
            + (index_score["energy"] * score_weights.energy)
        )

    raw_df = pd.DataFrame.from_records(index_scores).set_index("frame_index")

    # Normalize the score columns
    scaler = preprocessing.MinMaxScaler()
    scaled_df = pd.DataFrame(
        scaler.fit_transform(raw_df), columns=raw_df.columns, index=raw_df.index
    )

    scaled_df["entropy"] = 1.0 - scaled_df["entropy"]  # lower entropy = better

    scaled_df["overall_score"] = scaled_df.apply(_basic_score, axis=1)

    winning_indices = _apply_radius_deselection(
        indices_sorted_by_score=scaled_df["overall_score"]
        .sort_values(ascending=False)
        .index.to_numpy(),
        num_output_frames=num_output_frames,
        initial_radius=deselection_radius_frames,
    )

    if plot_path is not None:
        score_axes, other_axes = scaled_df.plot(
            kind="line",
            subplots=[("overall_score",), ("entropy", "variance", "saliency", "energy")],
            title="Frame Scores",
            xlabel="Frame Index",
            ylabel="Score",
            sharey=True,
            figsize=(12, 6),
            grid=True,
        )

        score_axes.set_title(label="Overall Score", fontsize="small")
        score_axes.legend().set_visible(False)

        other_axes.set_title(label="Score Composition", fontsize="small")

        for i in winning_indices:
            score_axes.axvline(i, color="lime", alpha=0.5)

        plt.tight_layout()
        plt.savefig(plot_path)

    return ScoredFrames(
        best_frame_index=int(scaled_df["overall_score"].idxmax()), winning_indices=winning_indices
    )


def discover_high_scoring_frames(  # pylint: disable=too-many-positional-arguments
    scoring_frames: ImageSourceType,
    intermediate_info: Optional[IntermediateFileInfo],
    batch_size: int,
    source_frame_count: int,
    conversion_scoring_functions: ConversionScoringFunctions,
    gpus: Tuple[GPUDescription, ...],
    num_output_frames: int,
    deselection_radius_frames: int,
    plot_path: Optional[Path],
) -> ScoredFrames:
    """
    Function to only get the list of high scoring frames in the input.

    :param scoring_frames: These frames are fed to the GPU via the conversion function and will
    likely be scaled down. Pre-scaled and buffered images should be passed here to avoid waiting
    around.
    :param source_frame_count: Number of frames in the input.
    :param intermediate_info: If given, vectors will be written to disk for faster
    re-runs.
    :param num_output_frames: Number of frames to select.
    :param batch_size: Number of scaled input images to send to the GPU at once.
    :param conversion_scoring_functions: Functions to compute vectors then score them.
    :param deselection_radius_frames: Frames surrounding high scoring frames removed to
    prevent clustering. This is the number of frames before/after a high scoring one that are
    slightly decreased in score.
    :param gpus: Describes the GPUs we're allowed to use for this computation.
    :param plot_path: If given, a visualization of the math that went into scores will be written
    to this path.
    :return:
    """

    vectors: Iterator[npt.NDArray[np.float16]] = conversion.frames_to_vectors(
        frames=scoring_frames,
        intermediate_info=intermediate_info,
        batch_size=batch_size,
        total_input_frames=source_frame_count,
        convert_batches=conversion_scoring_functions.conversion,
        gpus=gpus,
    )

    LOGGER.debug("Starting to sort output vectors by score.")

    score_indexes: List[IndexScores] = list(
        map(
            conversion_scoring_functions.scoring,
            enumerate(vectors),
        )
    )

    scored_sorted = _score_and_sort_frames(
        score_weights=conversion_scoring_functions.weights,
        index_scores=score_indexes,
        num_output_frames=num_output_frames,
        deselection_radius_frames=deselection_radius_frames,
        plot_path=plot_path,
    )

    return scored_sorted


def reduce_frames_by_score(  # pylint: disable=too-many-arguments, too-many-positional-arguments, too-many-locals
    scoring_frames: ImageSourceType,
    output_frames: ImageSourceType,
    source_frame_count: int,
    intermediate_info: Optional[IntermediateFileInfo],
    output_path: Path,
    num_output_frames: int,
    output_fps: float,
    batch_size: int,
    conversion_scoring_functions: ConversionScoringFunctions,
    deselection_radius_frames: int,
    gpus: Tuple[GPUDescription, ...],
    audio_paths: List[Path],
    plot_path: Optional[Path],
    best_frame_path: Optional[Path],
) -> ImageSourceType:
    """
    Uses input scoring function to reduce the input frames to the desired count by selecting
    the highest scoring frames.

    :param scoring_frames: These frames are fed to the GPU via the conversion function and will
    likely be scaled down. Pre-scaled and buffered images should be passed here to avoid waiting
    around.
    :param output_frames: These are the frames that are actually cropped and forwarded in the
    output.
    :param source_frame_count: Number of frames in the input.
    :param intermediate_info: If given, vectors will be written to disk for faster
    re-runs.
    :param output_path: Desired output location of the video.
    :param num_output_frames: Number of frames to select.
    :param output_fps: Set as the FPS of the output video.
    :param batch_size: Number of scaled input images to send to the GPU at once.
    :param conversion_scoring_functions: Functions to compute vectors then score them.
    :param deselection_radius_frames: Frames surrounding high scoring frames removed to
    prevent clustering. This is the number of frames before/after a high scoring one that are
    slightly decreased in score.
    :param gpus: Describes the GPUs we're allowed to use for this computation.
    :param audio_paths: If given, the audio files will be written to the output video.
    :param plot_path: If given, a visualization of the math that went into scores will be written
    to this path.
    :param best_frame_path: If given, the highest scoring frame in the input will be written to
    this path. Good for thumbnails.
    :return: Selected frames.
    """

    scored_sorted = discover_high_scoring_frames(
        scoring_frames=scoring_frames,
        intermediate_info=intermediate_info,
        batch_size=batch_size,
        source_frame_count=source_frame_count,
        conversion_scoring_functions=conversion_scoring_functions,
        gpus=gpus,
        num_output_frames=num_output_frames,
        deselection_radius_frames=deselection_radius_frames,
        plot_path=plot_path,
    )

    most_interesting_indices: Set[int] = set(sorted(scored_sorted.winning_indices))

    final_frame_index: int = max(most_interesting_indices)

    # Slice the frames to include the frame at final_frame_index
    # Ensure the frame at final_frame_index is included, the plus one.
    sliced_frames: ImageSourceType = itertools.islice(output_frames, None, final_frame_index + 1)

    def filter_frame(index_frame: Tuple[int, RGBInt8ImageType]) -> bool:
        """
        Hijacks the enumeration of the output frames to write the best frame.
        :param index_frame: Indices and the contents of the frames.
        :return: Filter output, True if the frame is interesting False if otherwise.
        """

        frame_index, frame = index_frame

        if frame_index == scored_sorted.best_frame_index and best_frame_path is not None:
            image_common.save_rgb_image(path=best_frame_path, image=frame)

        return frame_index in most_interesting_indices

    most_interesting_frames: ImageSourceType = (
        index_frame[1]
        for index_frame in filter(
            filter_frame,
            enumerate(sliced_frames),
        )
    )

    return video_common.write_source_to_disk_forward(
        source=most_interesting_frames,
        video_path=output_path,
        video_fps=output_fps,
        audio_paths=audio_paths,
        high_quality=True,
    )



"""
Combines the HTML templating process with image manipulation, enabling drawing webpages on top of
images.

Note: Needs google chrome!!

wget https://dl.google.com/linux/direct/google-chrome-stable_current_amd64.deb
sudo apt install ./google-chrome-stable_current_amd64.deb

Eventually we should use a docker image or something to do this.
"""

import tempfile
from pathlib import Path

import numpy as np
from html2image import Html2Image
from PIL import Image

from content_aware_timelapse.viderator import image_common
from content_aware_timelapse.viderator.html_on_image import read_html_template
from content_aware_timelapse.viderator.html_on_image.read_html_template import FilledTemplate
from content_aware_timelapse.viderator.viderator_types import RGBInt8ImageType


def template_over_image(
    filled_template: FilledTemplate, image: RGBInt8ImageType
) -> RGBInt8ImageType:
    """
    Take a filled template document (a string that is an HTML document), render it as an image,
    and composite that image on top of the input `image`.
    :param filled_template: HTML text.
    :param image: To draw on.
    :return: Document rendered and added on top of `image`.
    """

    input_resolution = image_common.image_resolution(image)

    with tempfile.TemporaryDirectory() as temp_dir:

        Path(temp_dir).mkdir(parents=True, exist_ok=True)

        renderer = Html2Image(
            size=(input_resolution.width, input_resolution.height),
            custom_flags=[
                "--hide-scrollbars",
                "--force-device-scale-factor=1.0",
                "--default-background-color=00000000",
            ],
            output_path=temp_dir,
            browser_executable="/usr/bin/google-chrome",
        )

        renderer.screenshot(html_str=filled_template, save_as="overlay.png")
        overlay = Image.open(Path(temp_dir) / "overlay.png").convert("RGBA")

    input_base = Image.fromarray(image).convert("RGBA")
    input_base.alpha_composite(overlay, (0, 0))

    # Convert back to RGB, squashing alpha, and then to numpy array
    final_array = np.array(input_base.convert("RGB"), dtype=np.uint8)

    return RGBInt8ImageType(final_array)


def create_simple_thumbnail(  # pylint: disable=too-many-positional-arguments
    image: RGBInt8ImageType,
    upper_subtitle: str,
    main_title: str,
    lower_title: str,
    gradient_start: str,
    gradient_stop: str,
    text_color: str,
    shadow_color: str,
) -> RGBInt8ImageType:
    """
    Convenience function, combining the "simple thumbnail overlay" asset with a given image.
    :param image: Rendered HTML document is composited over this image.
    :param upper_subtitle: Passed to template.
    :param main_title: Passed to template.
    :param lower_title: Passed to template.
    :param gradient_start: Passed to template.
    :param gradient_stop: Passed to template.
    :param text_color: Passed to template.
    :param shadow_color: Passed to template.
    :return: The input image with the rendered "Simple Thumbnail Overlay" drawn on top.
    """

    input_resolution = image_common.image_resolution(image)

    filled_template = read_html_template.simple_thumbnail_overlay(
        width=input_resolution.width,
        height=input_resolution.height,
        upper_subtitle=upper_subtitle,
        main_title=main_title,
        lower_title=lower_title,
        gradient_start=gradient_start,
        gradient_stop=gradient_stop,
        text_color=text_color,
        shadow_color=shadow_color,
    )

    return template_over_image(
        filled_template=filled_template,
        image=image,
    )



"""
Types needed to describe the frames to vectors conversions.
"""

from typing import Iterator, List, NamedTuple, Optional, Protocol, Tuple, TypedDict

import numpy as np
import numpy.typing as npt

from content_aware_timelapse.gpu_discovery import GPUDescription
from content_aware_timelapse.viderator.viderator_types import (
    ImageResolution,
    RGBInt8ImageType,
    XYPoint,
)


class IndexScores(TypedDict):
    """
    Links the frame index of the input with a set of overall scores that describe the whole frame
    """

    frame_index: int
    entropy: float
    variance: float
    saliency: float
    energy: float


class ConvertBatchesFunction(Protocol):
    """
    Describes functions that convert a batch of frames to vectors. This way multiple converters
    can be substituted.
    """

    def __call__(
        self, frame_batches: Iterator[List[RGBInt8ImageType]], gpus: Tuple[GPUDescription, ...]
    ) -> Iterator[npt.NDArray[np.float16]]:
        """
        :param frame_batches: Iterator of batches (lists) of frames for conversion.
        :param gpus: List of ids of the GPUs to use for conversion. Eventually this can be prompted
        to something more descriptive than an int if need be.
        :return: An iterator of the converted vectors.
        """


class ScoreVectorsFunction(Protocol):
    """
    Describes functions that convert vectors to numerical properties of the vectors.
    """

    def __call__(self, packed: Tuple[int, npt.NDArray[np.float16]]) -> IndexScores:
        """
        :param packed: A tuple, the index of the frame in the input and the calculated vectors
        for that frame.
        :return: An IndexScores, which are the numerical properties of the vectors.
        """


class ScoreWeights(NamedTuple):
    """
    Each of these values is a float between 0 and 1, and it is a multipler on the given column
    post normalization.
    """

    low_entropy: float
    variance: float
    saliency: float
    energy: float


class ConversionScoringFunctions(NamedTuple):
    """
    Links a conversion function (which goes from images -> vectors) to a scoring function which
    determines the numerical score for each vector. These two components make up both halves of
    the images to scores pipeline.
    """

    name: str
    conversion: ConvertBatchesFunction
    scoring: ScoreVectorsFunction
    weights: ScoreWeights
    max_side_length: Optional[int] = None
    """
    Allows pre-processing steps to shrink the input images to the exact side length, or nearby
    to decrease overall memory pressure, or pack more frames into memory before being processed.
    """


class IndexPointsOfInterest(TypedDict):
    """
    Links the frame index of the input with a list of points of interest within the image.
    This is consumed in auto-cropping to find the best region of the video.
    """

    frame_index: int
    points_of_interest: List[XYPoint]


class ScorePOIsFunction(Protocol):
    """
    Describes functions that convert vectors to points of interest in the source images.
    """

    def __call__(
        self,
        packed: Tuple[int, npt.NDArray[np.float16]],
        original_source_resolution: ImageResolution,
        num_interesting_points: int,
    ) -> IndexPointsOfInterest:
        """
        :param packed: A tuple, the index of the frame in the input and the vectors (an attention
        map) for that frame.
        :param original_source_resolution: Used to remap interesting locations back onto the full
        coordinate space of the source image.
        :param num_interesting_points: The number of points to generate per frame.
        :return: An NT that links the frame index with the list of interesting points.
        """


class ConversionPOIsFunctions(NamedTuple):
    """
    Similar to `ConversionScoringFunctions`, links the function that generates tensor
    with a function that creates the POIs.
    """

    name: str
    conversion: ConvertBatchesFunction
    compute_pois: ScorePOIsFunction
    max_side_length: Optional[int] = None
    """
    Allows pre-processing steps to shrink the input images to the exact side length, or nearby
    to decrease overall memory pressure, or pack more frames into memory before being processed.
    """



"""
For tracking how long things take.
"""

import time
from typing import Callable, ParamSpec, Tuple, TypeVar

T = TypeVar("T")


P = ParamSpec("P")


def measure_execution_time(
    func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> Tuple[T, float]:
    """
    Measures the execution time of a function and returns both the time taken and the result.
    :param func: A callable function to measure.
    :param args: Positional arguments to forward to the function.
    :param kwargs: Keyword arguments to forward to the function.
    :return: A tuple containing the execution time in seconds and the result of the function.
    """
    start_time = time.perf_counter()
    result = func(*args, **kwargs)  # Forward args and kwargs to the original function
    end_time = time.perf_counter()
    time_taken = end_time - start_time
    return result, time_taken


def measure_execution_time_decorator(func: Callable[P, T]) -> Callable[P, Tuple[T, float]]:
    """
    A decorator that measures the execution time of a function and returns
    both the time taken and the result of the function.
    :param func: A callable function to measure.
    :return: A function that returns a tuple of execution time and result.
    """

    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Tuple[T, float]:
        return measure_execution_time(func, *args, **kwargs)  # Call the helper function

    return wrapper



"""
Common types used in image/video
"""

from typing import Iterator, List, NamedTuple, NewType, Optional

import click
import numpy as np
from numpy import typing as npt
from PIL.Image import Image as PILImage  # pylint: disable=unused-import

# Other modules can/should import the standard PILImage type from here.

RGBInt8ImageType = NewType("RGBInt8ImageType", npt.NDArray[np.uint8])
"""
Dimensions are (Width, Height, Colors). Type is np.uint8
"""

ImageSourceType = Iterator[RGBInt8ImageType]


class ImageResolution(NamedTuple):
    """
    Standard NT for image dimensions. Creators are responsible for making sure the order is
    correct.
    """

    width: int
    height: int


class XYPoint(NamedTuple):
    """
    Standard NT for pixel locations. (0, 0) is in the top left corner of the image.
    """

    x: int
    y: int


class AspectRatio(NamedTuple):
    """
    Standard NT for an aspect ratio.
    """

    width: float
    height: float


class RectangleRegion(NamedTuple):
    """
    Standard NT for defining a rectangular area of an image.
    """

    top: int
    left: int
    bottom: int
    right: int


class AspectRatioParamType(click.ParamType):
    """
    Parameter for passing in an aspect ratio.
    """

    name: str = "aspect-ratio"

    def convert(
        self,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> AspectRatio:
        """
        Converts input string to namedtuple.
        :param value: To convert.
        :param param: Only consumed in error state.
        :param ctx: Only consumed in error state.
        :return: Parsed type.
        """

        try:
            width_str, height_str = str(value).split(":")
            width = float(width_str)
            height = float(height_str)
        except ValueError:
            self.fail(
                f"{value!r} is not a valid aspect ratio. Use the format WIDTH:HEIGHT",
                param,
                ctx,
            )

        if width <= 0 or height <= 0:
            self.fail(
                f"Aspect ratio values must be positive (got {width}:{height})",
                param,
                ctx,
            )

        return AspectRatio(width=width, height=height)


class UniqueIntMatrix2DParamType(click.ParamType):
    """
    Parameter type for a 2D matrix of unique integers.
    Accepts strings in the format '1,2;3,4' where rows are separated by ';' and
    columns by ','.
    """

    name = "unique-int-matrix-2d"

    def convert(
        self,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> List[List[int]]:
        """
        Convert input string to a 2D list of unique integers.
        """
        try:
            rows = value.split(";")
            matrix: List[List[int]] = []
            seen: set[int] = set()

            for row in rows:
                if not row.strip():
                    continue  # skip empty rows
                cols = []
                for x in row.split(","):
                    x_int = int(x.strip())
                    if x_int in seen:
                        raise ValueError(f"Duplicate value detected: {x_int}")
                    seen.add(x_int)
                    cols.append(x_int)
                matrix.append(cols)

            if not matrix:
                raise ValueError("Matrix cannot be empty")

            # Ensure all rows have the same number of columns
            n_cols = len(matrix[0])
            if any(len(r) != n_cols for r in matrix):
                raise ValueError("All rows must have the same number of columns")

            return matrix
        except Exception as e:  # pylint: disable=broad-except
            self.fail(f"{value!r} is not a valid 2D unique integer matrix: {e}", param, ctx)


class ImageResolutionParamType(click.ParamType):
    """
    Parameter for passing in an image resolution, e.g. '1920x1080'.
    """

    name: str = "WIDTHxHEIGHT"

    def convert(
        self,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> ImageResolution:
        """
        Converts 'WIDTHxHEIGHT' (case-insensitive) into an ImageResolution NT.
        """

        try:
            # Normalize separator
            width_str, height_str = str(value).lower().split("x")
            width = int(width_str)
            height = int(height_str)
        except Exception:  # pylint: disable=broad-except
            self.fail(
                f"{value!r} is not a valid image resolution. Use the format WIDTHxHEIGHT",
                param,
                ctx,
            )

        if width <= 0 or height <= 0:
            self.fail(
                f"Resolution values must be positive integers (got {width}x{height})",
                param,
                ctx,
            )

        return ImageResolution(width=width, height=height)



"""
Library code related to points of interest within inamges.
"""

from functools import partial
from typing import Iterator, List, NamedTuple, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numpy import typing as npt

from content_aware_timelapse.frames_to_vectors import conversion
from content_aware_timelapse.frames_to_vectors.conversion import IntermediateFileInfo
from content_aware_timelapse.frames_to_vectors.conversion_types import (
    ConversionPOIsFunctions,
    IndexPointsOfInterest,
)
from content_aware_timelapse.gpu_discovery import GPUDescription
from content_aware_timelapse.viderator.viderator_types import (
    ImageResolution,
    ImageSourceType,
    RectangleRegion,
    XYPoint,
)


class _PointFrameCount(NamedTuple):
    """
    Dynamic point with its frequency across frames.
    """

    point: XYPoint
    frame_count: int


def _count_frames_filter(
    points_of_interest: List[IndexPointsOfInterest],
    drop_frame_threshold: float,
) -> List[_PointFrameCount]:
    """
    Process the points of interest list, converting it to a list of points and the number of
    frames where that point appears.
    :param points_of_interest: A list of the POIs discovered for each input frame.
    :param drop_frame_threshold: Drop points that appear in more than this percent of frames.
    Float from 0-1
    :return: A list of points mapped to the number of frames they appear in.
    """

    # Flatten to DataFrame
    df = pd.DataFrame.from_records(
        (
            (pt.x, pt.y, frame["frame_index"])
            for frame in points_of_interest
            for pt in frame["points_of_interest"]
        ),
        columns=["x", "y", "frame_index"],
    )

    # Count in how many frames each (x, y) appears
    frames_per_point = (
        df.groupby(["x", "y"])["frame_index"].nunique().reset_index(name="frame_count")
    )

    # Keep only dynamic points
    dynamic_points_df = frames_per_point[
        (frames_per_point["frame_count"] / len(points_of_interest)) < drop_frame_threshold
    ]

    # Convert to list of DynamicPoint
    result: List[_PointFrameCount] = [
        _PointFrameCount(XYPoint(row.x, row.y), int(row.frame_count))
        for row in dynamic_points_df.itertuples(index=False)
    ]

    return result


class _ScoredRegion(NamedTuple):
    """
    Links a region with its score
    """

    score: float
    region: RectangleRegion


def _score_region(
    integral_image: Union[
        npt.NDArray[np.float32],
        npt.NDArray[np.int32],
    ],
    region: RectangleRegion,
) -> float:
    """
    Compute the sum of values within a rectangular region of an integral image.

    :param integral_image: 2D cumulative sum array. This could be the sum of frame count, or the
    sum of bulk attention.
    :param region: The region to compute the sum of within the integral region.
    :return: The total sum of the values inside the window region.
    """

    total = integral_image[region.bottom - 1, region.right - 1]
    if region.top > 0:
        total -= integral_image[region.top - 1, region.right - 1]
    if region.left > 0:
        total -= integral_image[region.bottom - 1, region.left - 1]
    if region.top > 0 and region.left > 0:
        total += integral_image[region.top - 1, region.left - 1]

    return float(total)


def _iou(region_a: RectangleRegion, region_b: RectangleRegion) -> float:
    """Compute intersection-over-union between two regions."""
    inter_left = max(region_a.left, region_b.left)
    inter_top = max(region_a.top, region_b.top)
    inter_right = min(region_a.right, region_b.right)
    inter_bottom = min(region_a.bottom, region_b.bottom)

    if inter_right <= inter_left or inter_bottom <= inter_top:
        return 0.0

    inter_area = (inter_right - inter_left) * (inter_bottom - inter_top)
    area_a = (region_a.right - region_a.left) * (region_a.bottom - region_a.top)
    area_b = (region_b.right - region_b.left) * (region_b.bottom - region_b.top)
    union_area = area_a + area_b - inter_area
    return inter_area / union_area if union_area > 0 else 0.0


def _top_regions(  # pylint: disable=too-many-locals,too-many-arguments
    points: List[_PointFrameCount],
    image_size: ImageResolution,
    region_resolution: ImageResolution,
    num_regions: int,
    alpha_points_frames: float,
) -> List[RectangleRegion]:
    """
    Exhaustively find the top regions of size `region_resolution` anywhere in the image. Score each
    region as a weighted combination of:
      - normalized unique point count in the region
      - normalized sum of frame counts in the region

    Selection process:
      1. Start with the highest scoring region.
      2. Iteratively add regions, always preferring zero-overlap regions.
      3. If no zero-overlap regions remain, pick regions that minimize overlap while maximizing
      score.
      4. Guarantees up to `num_regions` regions are returned if candidates exist.

    :param points: Points of interest within the frame.
    :param image_size: Size of the image.
    :param region_resolution: Desired size of the output region.
    :param num_regions: Number of regions to return.
    :param alpha_points_frames: Relative weight of unique points vs frame counts.
    :return: List of top `num_regions` RectangleRegion objects.
    """

    if not points:
        return []

    # Rasterize points into frame-count and unique-point maps
    frame_count_map = np.zeros((image_size.height, image_size.width), dtype=np.float32)
    unique_map = np.zeros((image_size.height, image_size.width), dtype=np.int32)

    for p in points:
        if 0 <= p.point.x < image_size.width and 0 <= p.point.y < image_size.height:
            frame_count_map[p.point.y, p.point.x] += p.frame_count
            unique_map[p.point.y, p.point.x] = 1

    # Integral images for O(1) sum calculations
    integral_frame = frame_count_map.cumsum(axis=0).cumsum(axis=1)
    integral_unique = unique_map.cumsum(axis=0).cumsum(axis=1)

    # Compute raw scores for all possible regions
    region_scores: List[_ScoredRegion] = []
    for top in range(0, image_size.height - region_resolution.height + 1):
        for left in range(0, image_size.width - region_resolution.width + 1):
            region = RectangleRegion(
                top=top,
                left=left,
                bottom=top + region_resolution.height,
                right=left + region_resolution.width,
            )
            frame_sum = _score_region(integral_image=integral_frame, region=region)
            unique_sum = _score_region(integral_image=integral_unique, region=region)

            if frame_sum > 0 or unique_sum > 0:
                region_scores.append(
                    _ScoredRegion(
                        score=alpha_points_frames * unique_sum
                        + (1 - alpha_points_frames) * frame_sum,
                        region=region,
                    )
                )

    if not region_scores:
        return []

    remaining = sorted(region_scores, key=lambda r: r.score, reverse=True)

    # Pre-populate with the top-scoring region first
    selected_regions: List[RectangleRegion] = [remaining.pop(0).region]

    # Iteratively pick remaining regions
    while len(selected_regions) < num_regions and remaining:

        # Partition candidates into zero-overlap vs overlapping
        zero_overlap_indices = [
            i
            for i, r in enumerate(remaining)
            if all(_iou(r.region, s) == 0.0 for s in selected_regions)
        ]

        if zero_overlap_indices:
            # Pick highest scoring zero-overlap region
            best_idx = max(zero_overlap_indices, key=lambda i: remaining[i].score)
        else:
            # Pick region with the highest score penalized by max overlap
            best_idx = max(
                range(len(remaining)),
                key=lambda i: remaining[i].score
                * (1 - max((_iou(remaining[i].region, s) for s in selected_regions), default=0.0)),
            )

        selected_regions.append(remaining[best_idx].region)
        remaining.pop(best_idx)

    return selected_regions[:num_regions]


def discover_poi_regions(  # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals
    analysis_frames: ImageSourceType,
    intermediate_info: Optional[IntermediateFileInfo],
    batch_size: int,
    source_frame_count: int,
    conversion_pois_functions: ConversionPOIsFunctions,
    original_resolution: ImageResolution,
    crop_resolution: ImageResolution,
    gpus: Tuple[GPUDescription, ...],
    num_regions: int,
    alpha_points_frames: float = 0.8,
) -> Tuple[RectangleRegion, ...]:
    """
    Returns a tuple of the most interesting regions in a frame.

    :param analysis_frames: These frames are fed to the GPU via the conversion function and will
    likely be scaled down. Pre-scaled and buffered images should be passed here to avoid waiting
    around.
    :param intermediate_info: If given, attention map vectors will be written to disk for faster
    re-runs.
    :param batch_size: Number of scaled input images to send to the GPU at once.
    :param source_frame_count: Number of frames in the input.
    :param conversion_pois_functions: Conversion POIs functions to compute and derive POIs from
    input frames.
    :param original_resolution: Resolution of the source.
    :param crop_resolution: Desired resolution to crop to.
    :param num_regions:
    :param alpha_points_frames:
    :param gpus: GPUs to use in the computation.
    :return: Output NT. Contains the region, the cropped output iterator and the visualization
    iterator.
    """

    vectors_for_pois: Iterator[npt.NDArray[np.float16]] = conversion.frames_to_vectors(
        frames=analysis_frames,
        intermediate_info=intermediate_info,
        batch_size=batch_size,
        total_input_frames=source_frame_count,
        convert_batches=conversion_pois_functions.conversion,
        gpus=gpus,
    )

    points_of_interest: List[IndexPointsOfInterest] = list(
        map(
            partial(
                conversion_pois_functions.compute_pois,
                original_source_resolution=original_resolution,
                num_interesting_points=30,
            ),
            enumerate(vectors_for_pois),
        )
    )

    winning_points_and_counts: List[_PointFrameCount] = _count_frames_filter(
        points_of_interest=points_of_interest,
        drop_frame_threshold=0.7,
    )

    return tuple(
        _top_regions(
            points=winning_points_and_counts,
            image_size=original_resolution,
            region_resolution=crop_resolution,
            num_regions=num_regions,
            alpha_points_frames=alpha_points_frames,
        )
    )



"""
Uses clip embeddings as vectors.
"""

import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Callable, Iterator, List, Tuple, cast

import clip
import more_itertools
import numpy as np
import torch
from numpy import typing as npt
from PIL import Image

from content_aware_timelapse.frames_to_vectors.conversion_types import (
    ConversionScoringFunctions,
    IndexScores,
    ScoreWeights,
)
from content_aware_timelapse.frames_to_vectors.vector_computation import gpu_model_management
from content_aware_timelapse.gpu_discovery import GPUDescription
from content_aware_timelapse.viderator.viderator_types import RGBInt8ImageType

LOGGER = logging.getLogger(__name__)

logging.getLogger("timm").setLevel(logging.ERROR)


def _compute_vectors_clip(
    frame_batches: Iterator[List[RGBInt8ImageType]], gpus: Tuple[GPUDescription, ...]
) -> Iterator[npt.NDArray[np.float16]]:
    """
    Computes CLIP embeddings for input frames using GPU acceleration if available.

    :param frame_batches: Iterator of lists of frames to compute vectors.
    :return: Iterator of CLIP feature vectors, one per input frame.
    """

    LOGGER.debug(f"Detected {torch.cuda.device_count()} GPUs. Loading CLIP Model")

    def load_clip_model(
        gpu_index: int,
    ) -> Tuple[torch.nn.Module, Callable[[Image.Image], torch.Tensor]]:
        """
        Loads the CLIP model onto the specified GPU.

        :param gpu_index: Target GPU index.
        :return: The CLIP model.
        """
        LOGGER.debug(f"Loading CLIP model onto GPU index: {gpu_index}...")
        try:
            model, preprocess = clip.load("ViT-B/16", device=f"cuda:{gpu_index}")
            model.eval()
            return model, preprocess
        except RuntimeError:
            LOGGER.error("Error loading CLIP model!")
            raise

    # Load models onto each available GPU
    models = [load_clip_model(i) for i in gpus]

    def process_images_for_model(
        image_batch: List[RGBInt8ImageType],
        model: torch.nn.Module,
        preprocess: Callable[[Image.Image], torch.Tensor],
    ) -> torch.Tensor:
        """
        Preprocesses images and extracts CLIP feature vectors.

        :param image_batch: List of RGB images as NumPy arrays.
        :param model: The CLIP model.
        :param preprocess: The CLIP preprocessing pipeline.
        :return: Feature vector tensor.
        """
        LOGGER.debug("Processing images for CLIP inference.")

        # Convert images to PIL and preprocess
        tensor_batch = torch.stack([preprocess(Image.fromarray(img)) for img in image_batch])

        # Move to GPU
        device = next(model.parameters()).device
        tensor_batch = tensor_batch.to(device)

        with torch.no_grad():
            return cast(torch.Tensor, model.encode_image(tensor_batch).half())

    def images_to_feature_vectors(
        image_batches: List[List[RGBInt8ImageType]], executor: ThreadPoolExecutor
    ) -> Iterator[npt.NDArray[np.float16]]:
        """
        Converts batches of images into CLIP feature vectors using multiple GPUs.

        :param image_batches: List of batches of RGB images.
        :param executor: ThreadPoolExecutor for concurrent processing.
        :return: Iterator of feature vectors.
        """
        futures = [
            executor.submit(process_images_for_model, image_batch, model, preprocess)
            for (image_batch, (model, preprocess)) in zip(image_batches, models)
        ]

        for index, future in enumerate(as_completed(futures)):
            yield from future.result().cpu().numpy()
            LOGGER.debug(f"Processed image batch #{index} using CLIP.")

    with gpu_model_management.gpu_cleanup(models=[m_f[0] for m_f in models]):
        # Create a single thread pool for all image batches
        # GPU parallel consumption is achieved in a hacky way here, by splitting the input batches
        # into n lists where n is the number of GPUS, then giving each CPU and sub list to a thread.
        # Thread pool for parallel GPU processing
        with ThreadPoolExecutor(max_workers=len(models)) as e:
            for batch in more_itertools.chunked(frame_batches, len(models)):
                yield from images_to_feature_vectors(batch, e)


def _calculate_scores_clip(packed: Tuple[int, npt.NDArray[np.float16]]) -> IndexScores:
    """
    Calculate a combined score from various metrics of the CLIP embedding.

    Metrics:
        - Entropy: Measures how distributed the embedding values are.
        - Variance: Measures spread in embedding values.
        - Saliency: Measures maximum embedding value.
        - Energy: Measures total embedding magnitude (L2 norm).

    :param packed: Tuple of the index and the CLIP embedding vector.
    :return: Combined score and index.
    """

    index, embedding = packed
    embedding = embedding.astype(np.float32)  # Convert to higher precision for calculations

    # Normalize embedding
    norm_embedding = embedding / (np.linalg.norm(embedding) + 1e-8)

    entropy_score = -np.sum(
        norm_embedding[norm_embedding > 0] * np.log(norm_embedding[norm_embedding > 0] + 1e-8)
    )

    return IndexScores(
        frame_index=index,
        entropy=entropy_score,
        variance=float(np.var(embedding)),
        saliency=float(np.max(embedding)),
        energy=float(np.linalg.norm(embedding)),
    )


CONVERT_CLIP = ConversionScoringFunctions(
    name="clip",
    conversion=_compute_vectors_clip,
    scoring=_calculate_scores_clip,
    weights=ScoreWeights(low_entropy=1, variance=1, saliency=1, energy=1),
)



"""
Contains the frames to vectors pipeline, the actual conversion functions are standalone modules.
"""

import itertools
import logging
from pathlib import Path
from typing import Iterator, NamedTuple, Optional, Tuple

import more_itertools
import numpy as np
from numpy import typing as npt

from content_aware_timelapse import vector_file
from content_aware_timelapse.frames_to_vectors.conversion_types import ConvertBatchesFunction
from content_aware_timelapse.gpu_discovery import GPUDescription
from content_aware_timelapse.viderator.viderator_types import ImageSourceType

LOGGER = logging.getLogger(__name__)


class IntermediateFileInfo(NamedTuple):
    """
    Fields needed to define an intermediate vector file.
    """

    path: Path
    signature: str


def frames_to_vectors(  # pylint:disable=too-many-positional-arguments
    frames: ImageSourceType,
    intermediate_info: Optional[IntermediateFileInfo],
    batch_size: int,
    total_input_frames: int,
    convert_batches: ConvertBatchesFunction,
    gpus: Tuple[GPUDescription, ...],
) -> Iterator[npt.NDArray[np.float16]]:
    """
    Computes feature vectors from an input iterator or frames.
    Because this process is expensive, even with GPU, the intermediate vectors are written to disk
    to avoid re-doing the work.

    TODO This is the function we want to benchmark.
    :param frames: Frames to process.
    :param intermediate_info: Describes the intermediate vector file, if not given it will not
    be used.
    :param batch_size: Number of frames to process at once. Should try to utilize all GPU memory.
    :param total_input_frames: Number of frames in `frames`.
    :param convert_batches: Function for going from a batch of images to vectors.
    :param gpus: Tuple of GPU ids to use.
    :return: Vectors, one per input frame.
    """

    if intermediate_info is not None:

        intermediate = vector_file.read_vector_file(
            vector_file=intermediate_info.path, input_signature=intermediate_info.signature
        )

        LOGGER.info(f"Read in {intermediate.length} intermediate vectors from file.")

        fresh_tensors: Iterator[npt.NDArray[np.float16]] = iter([])

        if intermediate.length < total_input_frames:

            LOGGER.info(f"Need to compute {total_input_frames-intermediate.length} new vectors.")

            # Skip to the unprocessed section of the input.
            unprocessed_frames: ImageSourceType = itertools.islice(
                frames, intermediate.length, None
            )

            # Compute new vectors, writing the results to disk.
            fresh_tensors = convert_batches(
                frame_batches=more_itertools.chunked(unprocessed_frames, batch_size),
                gpus=gpus,
            )

            fresh_tensors = vector_file.write_vector_file_forward(
                vector_iterator=fresh_tensors,
                vector_file=intermediate_info.path,
                input_signature=intermediate_info.signature,
            )

        yield from itertools.chain.from_iterable(
            (
                intermediate.iterator,  # first output any vectors from disk.
                fresh_tensors,  # second, compute any vectors not found on disk.
            )
        )

    else:
        yield from convert_batches(
            frame_batches=more_itertools.chunked(frames, batch_size),
            gpus=gpus,
        )



"""
Because computing vectors can only happen with GPUs, which is expensive, we want to cache the
results of these operations on disk as hdf5 files, known within the application as "vector files".

This module relates to reading/writing these files.
"""

import hashlib
import io
import itertools
import json
import logging
import re
from pathlib import Path
from typing import Dict, Iterator, List, NamedTuple, Optional, Tuple

import h5py
import numpy as np
from numpy import typing as npt

LOGGER = logging.getLogger(__name__)

SIGNATURE_ATTRIBUTE_NAME = "signature"
VERSION_ATTRIBUTE_NAME = "vector_file_version"
CURRENT_VECTOR_FILE_VERSION = "1.0.0"


def create_videos_signature(video_paths: List[Path], modifications_salt: Optional[str]) -> str:
    """
    Used to describe the contents of a vector file. Becomes an attribute in the HDF5 file.
    :param video_paths: Paths to the videos in the vector file.
    :param modifications_salt: Added to the output signature. Should be enough to distinguish two
    copies of the same input videos that have been modified.
    :return: A string, ready to be input to the HDF5 file.
    """

    def compute_partial_hash(video_path: Path) -> str:
        """
        Hashes the first 512KB of a video file.
        :param video_path: Path to the video.
        :return: Hash digest as a string.
        """

        sha256 = hashlib.sha256()

        with video_path.open("rb") as f:
            sha256.update(f.read(512_000))

        return sha256.hexdigest()

    videos_list: List[Dict[str, str]] = [
        {
            video.name: json.dumps(
                {
                    "file_size": video.stat().st_size,
                    "partial_sha256": compute_partial_hash(video),
                }
            )
            for video in sorted(video_paths, key=str)
        }
    ]

    # Convert dictionary to JSON string
    return json.dumps(
        {
            "modifications_salt": modifications_salt,
            "videos": videos_list,
        }
    )


class _LengthIterator(NamedTuple):
    """
    Intermediate type.
    Links the count of vectors on disk, in the file with an iterator that will emit those vectors.
    """

    length: int
    iterator: Iterator[npt.NDArray[np.float16]]


def _sort_groups_by_index(group_names: List[str]) -> List[str]:
    """
    Sort frame groups by the integer part of their names to preserve the correct order.
    Assumes frame names are in the format of "frame_<index>", where index is an integer.

    :param group_names: List of group names (strings) to sort.
    :return: List of group names sorted by their integer index.
    """
    # Extract the integer part from the frame names and sort by it
    return sorted(group_names, key=lambda x: int(re.search(r"(\d+)", x).group(1)))


def read_vector_file(vector_file: Path, input_signature: str) -> _LengthIterator:
    """
    Reads vectors from an HDF5 "vector file" as an iterator. When the iterator in the output is
    exhausted, it is auto-closed.

    If the input file is empty, the resulting count will be zero and the iterator will have
    no items in it.

    :param vector_file: Path to the HDF5 file.
    :param input_signature: Signature to validate against.
    :return: NT containing the number of vectors on disk and an iterator that produces them.
    """

    f = h5py.File(vector_file, "a")

    def iterate_from_file() -> Iterator[npt.NDArray[np.float16]]:
        """
        Yields each dataset in the frame groups.
        :return: An iterator of the datasets.
        """
        try:
            # Sort the group names using _sort_groups_by_index to preserve integer order
            for frame_name in _sort_groups_by_index(group_names=list(f.keys())):
                frame_group = f[frame_name]

                for dataset_name in frame_group:
                    dataset = frame_group[dataset_name]

                    # Ensure the dtype matches
                    yield dataset[()]
        finally:
            # Automatically close the file when iteration is done.
            f.close()

    # Validate the signature and check for frame groups
    if (
        SIGNATURE_ATTRIBUTE_NAME in f.attrs.keys()
        and f.attrs[SIGNATURE_ATTRIBUTE_NAME] == input_signature
        and len(f.keys())  # Ensure there are frame groups present
    ):
        completed_vectors = sum(
            len(f[frame_group].keys())
            for frame_group in f.keys()  # pylint: disable=consider-using-dict-items
        )

        return _LengthIterator(
            length=completed_vectors,
            iterator=iterate_from_file(),
        )
    else:
        # Close the file if the signature doesn't match or the file is empty
        f.close()
        return _LengthIterator(
            length=0,
            iterator=iter([]),
        )


def _setup_vector_file_metadata(f: h5py.File, input_signature: str) -> int:
    """
    Sets up the HDF5 file metadata for vector storage, ensuring compatibility with the input
    signature.

    :param f: HDF5 file object.
    :param input_signature: Signature string for validation or initialization.
    :return: The starting index for writing new frames.
    """
    # Determine starting index based on the existing frame groups
    if (
        SIGNATURE_ATTRIBUTE_NAME in f.attrs
        and f.attrs[SIGNATURE_ATTRIBUTE_NAME] == input_signature
        and len(f.keys()) > 0
    ):
        # Extract existing frame group names and find the highest index
        existing_indices = [
            int(frame_name.split("_")[1])
            for frame_name in f.keys()
            if frame_name.startswith("frame_")
        ]
        starting_index = max(existing_indices) + 1 if existing_indices else 0

    elif (
        SIGNATURE_ATTRIBUTE_NAME in f.attrs
        and f.attrs[SIGNATURE_ATTRIBUTE_NAME] != input_signature
        and len(f.keys()) > 0
    ):
        # If the signature doesn't match but the file is populated, raise an error
        raise ValueError("Can't write to vector file! Signature does not match.")

    else:
        # Initialize metadata for a new file
        f.attrs[SIGNATURE_ATTRIBUTE_NAME] = input_signature
        f.attrs[VERSION_ATTRIBUTE_NAME] = CURRENT_VECTOR_FILE_VERSION
        starting_index = 0

    return starting_index


def _create_hdf5_worker(
    index: str, vector: npt.NDArray[np.float16]
) -> Tuple[str, bytes, npt.NDArray[np.float16]]:
    """
    Create an in-memory HDF5 file of the input vector and return the BytesIO object containing that
    vector.
    :param index: The index of the frame of the input video that produced these vectors.
    :param vector: The vector to export.
    :return: A Tuple (The index forwarded from the input, the BytesIO containing the vector).
    """

    with io.BytesIO() as bytes_io:
        with h5py.File(bytes_io, "w") as f:
            f.create_dataset(
                name=f"vit_base_patch16_224_{index}",
                shape=vector.shape,
                dtype=vector.dtype,
                data=vector,
                compression="gzip",
                compression_opts=9,
                chunks=True,
                maxshape=vector.shape,
            )
            f.flush()

        output = index, bytes_io.getvalue(), vector

    return output


def write_vector_file_forward(
    vector_iterator: Iterator[npt.NDArray[np.float16]],
    vector_file: Path,
    input_signature: str,
) -> Iterator[npt.NDArray[np.float16]]:
    """
    Creates a new iterator that is a copy of the input iterator. Each time the new iterator is
    iterated upon, that file is written to disk in the input file.
    :param vector_iterator: Input vectors.
    :param vector_file: Path to the file to write.
    :param input_signature: Added to the newly created vector file as an attribute.
    :return: The input vectors in their original order. Now they've been written to disk.
    """

    with h5py.File(vector_file, "a") as f:

        starting_index = _setup_vector_file_metadata(f, input_signature)

        LOGGER.info(f"Starting to write vectors to: {vector_file}")

        indices: Iterator[str] = map(str, itertools.count(starting_index))

        results: Iterator[Tuple[str, npt.NDArray[np.float16]]] = zip(indices, vector_iterator)

        for index, vector in results:

            # New groups are created to be able to hold different datasets created for
            # each frame. For superstition reasons this also might prevent memory leaks.
            index_group = f.create_group(f"frame_{index}")

            index_group.create_dataset(
                name=f"vit_base_patch16_224_{index}",
                shape=vector.shape,
                dtype=vector.dtype,
                data=vector,
            )

            # Flushing is still recommended for safety
            f.flush()

            yield vector



"""Common GPU managment and usage code."""

import gc
from contextlib import contextmanager
from typing import Generator, Iterable, Optional

import torch


@contextmanager
def gpu_cleanup(models: Optional[Iterable[torch.nn.Module]] = None) -> Generator[None, None, None]:
    """
    Context manager that ensures GPU resources are released after use.

    Deletes provided models (if any) and clears the PyTorch CUDA cache,
    even if an exception occurs within the context.

    :param models: Iterable of torch.nn.Module objects to delete after use.
    :return: None
    """
    try:
        yield
    finally:
        if models is not None:
            # Explicitly free the models
            for m in models:
                del m
            gc.collect()
        # Clear any cached CUDA memory
        torch.cuda.empty_cache()
        torch.cuda.ipc_collect()  # type: ignore[no-untyped-call]



"""
Main functionality, defines the pipeline.
"""

import itertools
import json
import logging
import time
from functools import partial
from pathlib import Path
from statistics import mean
from typing import List, NamedTuple, Optional, Set, Tuple, cast

import more_itertools
from tqdm import tqdm

from assets import EASTERN_BOX_TURTLE_PATH
from content_aware_timelapse.frames_to_vectors import vector_scoring
from content_aware_timelapse.frames_to_vectors.conversion import IntermediateFileInfo
from content_aware_timelapse.frames_to_vectors.conversion_types import (
    ConversionPOIsFunctions,
    ConversionScoringFunctions,
)
from content_aware_timelapse.frames_to_vectors.vector_points_of_interest import discover_poi_regions
from content_aware_timelapse.frames_to_vectors.vector_scoring import (
    ScoredFrames,
    reduce_frames_by_score,
)
from content_aware_timelapse.gpu_discovery import GPUDescription
from content_aware_timelapse.vector_file import create_videos_signature
from content_aware_timelapse.viderator import (
    frames_in_video,
    image_common,
    iterator_common,
    video_common,
)
from content_aware_timelapse.viderator.video_common import VideoFrames
from content_aware_timelapse.viderator.viderator_types import (
    AspectRatio,
    ImageResolution,
    ImageSourceType,
    RectangleRegion,
)

LOGGER = logging.getLogger(__name__)


class _CombinedVideos(NamedTuple):
    """
    Intermediate type for keeping track of the total number of frames in an iterator.
    """

    total_frame_count: int
    frames: ImageSourceType
    original_resolution: ImageResolution
    original_fps: float


def calculate_output_frames(duration: float, output_fps: float) -> int:
    """
    Canonical function to do this math.
    :param duration: Desired length of the video in seconds.
    :param output_fps: FPS of output.
    :return: Round number for output frames.
    """

    return int(duration * output_fps)


def optionally_resize(
    output_resolution: Optional[ImageResolution], source: ImageSourceType
) -> ImageSourceType:
    """
    Helper function. If a desired output resolution is given, resize the source, otherwise just
    forward the source.
    :param output_resolution: Desired output resolution. Or not.
    :param source: To resize. Or not.
    :return: Resized Source. Or not.
    """

    if output_resolution is not None:
        return video_common.resize_source(
            source=source,
            resolution=output_resolution,
        )

    return source


def load_input_videos(
    input_files: List[Path],
    resize_inputs: bool,
    tqdm_desc: Optional[str] = None,
    tqdm_total: Optional[int] = None,
    take_only: Optional[int] = None,
) -> _CombinedVideos:
    """
    Helper function to combine the input videos.

    :param input_files: List of input videos.
    :param resize_inputs: If True and the input videos have different resolutions, the videos will
    be resized to the minimum resolution in the set.
    :param tqdm_desc: If provided, wrap the combined frame iterator in a tqdm progress bar
    with this description. If None, tqdm is disabled.
    :param tqdm_total: If provided, overrides the TQDM total field. By default, this is the total
    number of frames in the input video.
    :return: NT containing the total frame count and a joined iterator of all the input
    frames.
    """

    input_video_frames: List[VideoFrames] = list(
        map(frames_in_video.frames_in_video_opencv, input_files)
    )

    input_resolutions: Set[ImageResolution] = {
        video_frames.original_resolution for video_frames in input_video_frames
    }

    output_resolution: ImageResolution = next(iter(input_resolutions))

    if len(input_resolutions) > 1:

        if resize_inputs:

            aspect_ratios = set(list(map(image_common.resolution_aspect_ratio, input_resolutions)))

            if len(aspect_ratios) > 1:
                raise ValueError(
                    f"Cannot resize inputs with multiple input aspect ratios: {aspect_ratios}"
                )

            smallest_resolution: ImageResolution = min(
                input_resolutions, key=lambda resolution: resolution.width * resolution.height
            )

            output_resolution = smallest_resolution

            LOGGER.debug(f"Resizing all inputs to: {output_resolution}")

            all_input_frames: ImageSourceType = itertools.chain.from_iterable(
                video_common.resize_source(
                    source=video_frames.frames, resolution=smallest_resolution
                )
                for video_frames in input_video_frames
            )
        else:
            raise ValueError(f"Input videos have different resolutions: {input_video_frames}")
    else:
        all_input_frames = itertools.chain.from_iterable(
            video_frames.frames for video_frames in input_video_frames
        )

    if take_only is not None:
        all_input_frames = itertools.islice(all_input_frames, take_only)
        total_frames = take_only
    else:
        total_frames = sum(video_frames.total_frame_count for video_frames in input_video_frames)

    input_fps: Set[float] = {video_frames.original_fps for video_frames in input_video_frames}

    # Conditionally wrap in tqdm
    if tqdm_desc is not None:
        frames_iter = tqdm(
            all_input_frames,
            total=tqdm_total if tqdm_total is not None else total_frames,
            unit="Frames",
            ncols=100,
            desc=tqdm_desc,
            maxinterval=1,
        )
    else:
        frames_iter = all_input_frames

    return _CombinedVideos(
        total_frame_count=total_frames,
        frames=cast(ImageSourceType, frames_iter),
        original_resolution=output_resolution,
        original_fps=next(iter(input_fps)),
    )


def preload_and_scale(
    video_source: _CombinedVideos,
    max_side_length: Optional[int],
    buffer_size: int,
) -> _CombinedVideos:
    """
    Wraps a few library functions to optionally scale and pre-load input videos from disk into
    RAM for faster processing. See the underlying docs for `iterator_common.preload_into_memory`
    for the conditions of that load optimization.
    :param video_source: Source.
    :param max_side_length: If given, the source frames will be scaled such that maximum side
    length is scaled to this value while maintaining the source aspect ratio.
    :param buffer_size: Number of frames to preload into memory. If 0, no preloading will occur.
    :return: NT that switches the `frames` field of the input with the modified buffer result.
    """

    output_frames = video_source.frames

    if max_side_length is not None:
        output_frames = map(
            partial(
                image_common.resize_image_max_side,
                max_side_length=max_side_length,
                delete=True,
            ),
            output_frames,
        )

    if buffer_size > 0:
        output_frames = iterator_common.preload_into_memory(
            source=output_frames, buffer_size=buffer_size, fill_buffer_before_yield=True
        )

    return _CombinedVideos(
        frames=output_frames,
        total_frame_count=video_source.total_frame_count,
        original_resolution=video_source.original_resolution,
        original_fps=video_source.original_fps,
    )


def create_timelapse_score(  # pylint: disable=too-many-locals,too-many-positional-arguments,too-many-arguments
    input_files: List[Path],
    output_path: Path,
    duration: float,
    output_fps: float,
    output_resolution: Optional[ImageResolution],
    resize_inputs: bool,
    batch_size: int,
    buffer_size: int,
    conversion_scoring_functions: ConversionScoringFunctions,
    deselection_radius_frames: int,
    audio_paths: List[Path],
    gpus: Tuple[GPUDescription, ...],
    vectors_path: Optional[Path],
    plot_path: Optional[Path],
    best_frame_path: Optional[Path],
) -> None:
    """
    Library backend for the UI function. See docs in `reduce_frames_by_score` or docs in click
    CLI function for reference.

    :param input_files: See docs in library or click.
    :param output_path: See docs in library or click.
    :param duration: See docs in library or click.
    :param output_fps: See docs in library or click.
    :param output_resolution: See docs in library or click.
    :param resize_inputs: See docs in library or click.
    :param batch_size: See docs in library or click.
    :param buffer_size: See docs in library or click.
    :param conversion_scoring_functions: See docs in library or click.
    :param deselection_radius_frames: See docs in library or click.
    :param audio_paths: See docs in library or click.
    :param gpus: See docs in library or click.
    :param vectors_path: See docs in library or click.
    :param plot_path: See docs in library or click.
    :param best_frame_path: See docs in library or click.
    :return: None
    """

    frames_count_resolution = preload_and_scale(
        video_source=load_input_videos(
            input_files=input_files, tqdm_desc="Reading Score Frames", resize_inputs=resize_inputs
        ),
        max_side_length=conversion_scoring_functions.max_side_length,
        buffer_size=buffer_size,
    )

    output_image_source = reduce_frames_by_score(
        scoring_frames=frames_count_resolution.frames,
        output_frames=load_input_videos(
            input_files=input_files, tqdm_desc="Reading Output Frames", resize_inputs=resize_inputs
        ).frames,
        source_frame_count=frames_count_resolution.total_frame_count,
        intermediate_info=(
            IntermediateFileInfo(
                path=vectors_path,
                signature=create_videos_signature(video_paths=input_files, modifications_salt=None),
            )
            if vectors_path is not None
            else None
        ),
        output_path=output_path,
        num_output_frames=calculate_output_frames(duration=duration, output_fps=output_fps),
        output_fps=output_fps,
        batch_size=batch_size,
        conversion_scoring_functions=conversion_scoring_functions,
        deselection_radius_frames=deselection_radius_frames,
        audio_paths=audio_paths,
        plot_path=plot_path,
        gpus=gpus,
        best_frame_path=best_frame_path,
    )

    more_itertools.consume(
        optionally_resize(output_resolution=output_resolution, source=output_image_source)
    )


def create_timelapse_crop_score(  # pylint: disable=too-many-locals,too-many-positional-arguments,too-many-arguments,unused-argument,unused-variable
    input_files: List[Path],
    output_path: Path,
    duration: float,
    output_fps: float,
    output_resolution: Optional[ImageResolution],
    resize_inputs: bool,
    batch_size_pois: int,
    batch_size_scores: int,
    scaled_frames_buffer_size: int,
    conversion_pois_functions: ConversionPOIsFunctions,
    conversion_scoring_functions: ConversionScoringFunctions,
    scoring_deselection_radius_frames: int,
    audio_paths: List[Path],
    gpus: Tuple[GPUDescription, ...],
    layout_matrix: List[List[int]],
    pois_vectors_path: Optional[Path],
    scores_vectors_path: Optional[Path],
    plot_path: Optional[Path],
    aspect_ratio: Optional[AspectRatio] = None,
    crop_resolution: Optional[ImageResolution] = None,
) -> None:
    """
    Library backend for the UI function. See docs in `reduce_frames_by_score`, `crop_to_pois` or
    docs in click CLI function for complete reference.

    :param input_files: See docs in library or click.
    :param output_path: See docs in library or click.
    :param duration: See docs in library or click.
    :param output_fps: See docs in library or click.
    :param output_resolution: See docs in library or click.
    :param resize_inputs: See docs in library or click.
    :param batch_size_pois: See docs in library or click.
    :param batch_size_scores: See docs in library or click.
    :param scaled_frames_buffer_size: See docs in library or click.
    :param conversion_pois_functions: See docs in library or click.
    :param conversion_scoring_functions: See docs in library or click.
    :param aspect_ratio: See docs in library or click.
    :param crop_resolution: See docs in library or click.
    :param scoring_deselection_radius_frames: See docs in library or click.
    :param audio_paths: See docs in library or click.
    :param gpus: See docs in library or click.
    :param layout_matrix: See docs in library or click.
    :param pois_vectors_path: See docs in library or click.
    :param scores_vectors_path: See docs in library or click.
    :param plot_path: See docs in library or click.
    :return: None
    """

    if not any([aspect_ratio, crop_resolution]):
        raise ValueError("Either an aspect ratio or crop resolution must be provided.")

    num_regions: int = len(list(itertools.chain.from_iterable(layout_matrix)))

    poi_discovery_source: _CombinedVideos = load_input_videos(
        input_files=input_files,
        tqdm_desc="Reading Frames for POI Discovery",
        resize_inputs=resize_inputs,
    )

    scoring_source: ImageSourceType = load_input_videos(
        input_files=input_files,
        tqdm_desc="Reading Frames to Crop and Score",
        resize_inputs=resize_inputs,
    ).frames

    if crop_resolution is None:
        crop_resolution = image_common.largest_fitting_region(
            source_resolution=poi_discovery_source.original_resolution,
            aspect_ratio=aspect_ratio,
            even_dimensions=True,
        )

    poi_regions: Tuple[RectangleRegion, ...] = discover_poi_regions(
        analysis_frames=preload_and_scale(
            video_source=_CombinedVideos(
                frames=poi_discovery_source.frames,
                total_frame_count=poi_discovery_source.total_frame_count,
                original_resolution=poi_discovery_source.original_resolution,
                original_fps=poi_discovery_source.original_fps,
            ),
            max_side_length=conversion_pois_functions.max_side_length,
            buffer_size=scaled_frames_buffer_size,
        ).frames,
        intermediate_info=(
            IntermediateFileInfo(
                path=pois_vectors_path,
                signature=create_videos_signature(video_paths=input_files, modifications_salt=None),
            )
            if pois_vectors_path is not None
            else None
        ),
        batch_size=batch_size_pois,
        source_frame_count=poi_discovery_source.total_frame_count,
        conversion_pois_functions=conversion_pois_functions,
        original_resolution=poi_discovery_source.original_resolution,
        crop_resolution=crop_resolution,
        gpus=gpus,
        num_regions=num_regions,
    )

    cropped_to_best_region: ImageSourceType = video_common.crop_source(
        source=scoring_source,
        region=next(iter(poi_regions)),
    )

    buffered_scoring_frames: ImageSourceType = preload_and_scale(
        video_source=_CombinedVideos(
            frames=cropped_to_best_region,
            total_frame_count=poi_discovery_source.total_frame_count,
            original_resolution=crop_resolution,
            original_fps=poi_discovery_source.original_fps,
        ),
        max_side_length=conversion_scoring_functions.max_side_length,
        buffer_size=scaled_frames_buffer_size,
    ).frames

    scored_frames: ScoredFrames = vector_scoring.discover_high_scoring_frames(
        scoring_frames=buffered_scoring_frames,
        source_frame_count=poi_discovery_source.total_frame_count,
        intermediate_info=(
            IntermediateFileInfo(
                path=scores_vectors_path,
                signature=create_videos_signature(
                    video_paths=input_files,
                    modifications_salt=json.dumps({"winning_regions": list(poi_regions)}),
                ),
            )
            if scores_vectors_path is not None
            else None
        ),
        num_output_frames=calculate_output_frames(duration=duration, output_fps=output_fps),
        batch_size=batch_size_scores,
        conversion_scoring_functions=conversion_scoring_functions,
        deselection_radius_frames=scoring_deselection_radius_frames,
        plot_path=plot_path,
        gpus=gpus,
    )

    final_frame_index = max(scored_frames.winning_indices)

    output_source: ImageSourceType = itertools.islice(
        load_input_videos(
            input_files=input_files,
            tqdm_desc="Reading Frames for Crop/Score Output",
            tqdm_total=final_frame_index,
            resize_inputs=resize_inputs,
        ).frames,
        None,
        final_frame_index + 1,
    )

    def output_frames() -> ImageSourceType:
        """
        Creates the output frames.
        :return: Iterator of output frames to be written to disk.
        """

        for frame_index, input_frame in enumerate(output_source):
            if frame_index in scored_frames.winning_indices:
                yield image_common.reshape_from_regions(
                    image=input_frame,
                    prioritized_poi_regions=poi_regions,
                    layout_matrix=layout_matrix,
                )

    video_common.write_source_to_disk_consume(
        source=optionally_resize(output_resolution=output_resolution, source=output_frames()),
        video_path=output_path,
        video_fps=output_fps,
        audio_paths=audio_paths,
        high_quality=True,
    )


def benchmark(
    conversion_scoring_functions: ConversionScoringFunctions,
    gpus: Tuple[GPUDescription, ...],
    batch_size: int,
    batch_count: int,
    runs: int,
) -> float:
    """
    Package API counterpart of the CLI function.

    Returns the frames/per second the given set of GPUs can shove images through the conversion
    function. Helps measure improvements in the conversion function mechanics as well as evaluate
    performance of standalone CAT GPU nodes.

    1. A frame buffer of batch_size * batch_count + 1 is loaded into memory.
    2. One batch is converted to ensure all modeling loading etc has taken place.
    3. Start the timer.
    4. Remaining batches are converted, discarding the output.
    5. Timer stops.

    The result of (batch_size * batch_count) / duration in sections is reported.

    This whole process is done per the run value and the mean is reported.

    :param conversion_scoring_functions: Contains the function that converts batches of frames into
    vectors. Also, responsible for bootstrapping GPU environment. Only the conversion function
    and the side length are used in this function.
    :param gpus: To use in the benchmark.
    :param batch_size: Batch size processed at one time by the GPU.
    :param batch_count: Number of batches to process for the benchmark.
    :param runs: Number of times to run the benchmark.
    :return: Throughput in frames/second.
    """

    def single_run() -> float:
        """
        Run the benchmark once.
        :return: Frames per second as a float
        """

        full_size = image_common.load_rgb_image(path=EASTERN_BOX_TURTLE_PATH)
        shrunk_to_side_length = image_common.resize_image_max_side(
            image=full_size,
            max_side_length=conversion_scoring_functions.max_side_length,
            delete=True,
        )

        all_frames = [shrunk_to_side_length.copy() for _ in range(batch_size * (batch_count + 1))]

        output_iterator = conversion_scoring_functions.conversion(
            frame_batches=more_itertools.chunked(all_frames, batch_size),
            gpus=gpus,
        )

        _warmup_vectors = next(output_iterator)

        start = time.perf_counter()
        more_itertools.consume(output_iterator)
        end = time.perf_counter()

        # Free all references to the test images. Makes sure they're all gone on the GPU as well.
        del all_frames

        return (batch_size * batch_count) / (end - start)

    return mean([single_run() for _ in range(runs)])



"""
Utilities for finding GPUs attached to the system
"""

import logging
from typing import Tuple, TypeAlias

import nvsmi

GPUDescription: TypeAlias = int

LOGGER = logging.getLogger(__name__)


def discover_gpus() -> Tuple[GPUDescription, ...]:
    """
    Returns the currently visible GPUs.
    :return: A tuple of GPUs.
    """

    gpus = tuple(int(gpu.id) for gpu in nvsmi.get_gpus())
    LOGGER.debug(f"Discovered {len(gpus)} GPUs: {gpus}")
    return gpus



"""
Different functions to pull frames out of a video.
"""

import itertools
import logging
from fractions import Fraction
from pathlib import Path
from typing import Iterator, List, Optional, Protocol, cast

import cv2
import ffmpeg
import numpy as np

from content_aware_timelapse.viderator.video_common import VideoFrames, reduce_fps_take_every
from content_aware_timelapse.viderator.viderator_types import (
    ImageResolution,
    ImageSourceType,
    RGBInt8ImageType,
)

LOGGER = logging.getLogger(__name__)


class FramesInVideo(Protocol):
    """
    Protocol for different implementations of an RGB frame extractors.
    """

    def __call__(  # pylint: disable=too-many-positional-arguments
        self,
        video_path: Path,
        video_fps: Optional[float],
        reduce_fps_to: Optional[float],
        width_height: Optional[ImageResolution],
        starting_frame: Optional[int],
    ) -> VideoFrames:
        """
        :param video_path: The path to the video file on disk.
        :param video_fps: Can be used to override the actual FPS of the video.
        :param reduce_fps_to: Discards frames such that the frames that are returned are at this
        FPS.
        :param width_height: If given, the output frames will be resized to this resolution.
        :param starting_frame: Seek to this frame of the video open opening.
        :return: An NT containing metadata about the video, and an iterator that produces the
        frames. Frames are in RGB color order.
        :raises: ValueError if the video can't be opened, or the given `reduce_fps_to` is
        impossible.
        """


def frames_in_video_opencv(
    video_path: Path,
    video_fps: Optional[float] = None,
    reduce_fps_to: Optional[float] = None,
    width_height: Optional[ImageResolution] = None,
    starting_frame: Optional[int] = None,
) -> VideoFrames:
    """
    Reads frames from a file using openCV. CPU only.
    :param video_path: See docs in protocol.
    :param video_fps: See docs in protocol.
    :param reduce_fps_to: See docs in protocol.
    :param width_height: See docs in protocol.
    :param starting_frame: See docs in protocol.
    :return: See docs in protocol.
    """

    vid_capture = cv2.VideoCapture(str(video_path))

    file_fps = float(vid_capture.get(cv2.CAP_PROP_FPS))

    if video_fps:
        if video_fps != file_fps:
            LOGGER.warning(
                f"Override FPS of: {video_fps} fps "
                f"did not match the fps from the file of: {file_fps} fps."
            )
        fps = video_fps
    else:
        fps = file_fps

    frame_count = itertools.count()

    if starting_frame is not None:
        vid_capture.set(cv2.CAP_PROP_POS_FRAMES, starting_frame)
        frame_count = itertools.count(starting_frame)

    total_frame_count = int(vid_capture.get(cv2.CAP_PROP_FRAME_COUNT))

    take_every = reduce_fps_take_every(original_fps=fps, new_fps=reduce_fps_to)

    original_width_height = ImageResolution(
        width=int(vid_capture.get(cv2.CAP_PROP_FRAME_WIDTH)),
        height=int(vid_capture.get(cv2.CAP_PROP_FRAME_HEIGHT)),
    )

    resize = (original_width_height != width_height) if width_height is not None else False

    if not vid_capture.isOpened():
        raise ValueError(f"Couldn't open video file: {video_path}")

    def frames() -> ImageSourceType:
        """
        Read frames off of the video capture until there none left or pulling a frame fails.
        :return: An iterator of frames.
        """
        while vid_capture.isOpened():

            current_frame_index = next(frame_count)

            LOGGER.debug(
                f"Got a frame # {current_frame_index} / {total_frame_count} "
                f"({(current_frame_index / total_frame_count) * 100:.2f}) from: {video_path.name}"
            )

            ret, frame = vid_capture.read()
            if ret:
                image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                output = cast(
                    RGBInt8ImageType,
                    (
                        image
                        if not resize
                        else cv2.resize(image, (width_height.width, width_height.height))
                    ),
                )
                yield output
            else:
                break

    return VideoFrames(
        original_fps=vid_capture.get(cv2.CAP_PROP_FPS),
        original_resolution=original_width_height,
        total_frame_count=total_frame_count,
        frames=itertools.islice(frames(), None, None, take_every),
    )


def frames_in_video_ffmpeg(  # pylint: disable=too-many-locals
    video_path: Path,
    video_fps: Optional[float] = None,
    reduce_fps_to: Optional[float] = None,
    width_height: Optional[ImageResolution] = None,
    starting_frame: Optional[int] = None,
) -> VideoFrames:
    """
    Deprecated! Reads frames from a file using ffmpeg directly. CPU only.
    :param video_path: See docs in protocol.
    :param video_fps: See docs in protocol.
    :param reduce_fps_to: See docs in protocol.
    :param width_height: See docs in protocol.
    :param starting_frame: See docs in protocol.
    :return: See docs in protocol.
    """
    probe = ffmpeg.probe(str(video_path))
    video_stream = next(s for s in probe["streams"] if s["codec_type"] == "video")

    original_resolution = ImageResolution(
        width=video_stream["width"], height=video_stream["height"]
    )

    file_fps = float(Fraction(video_stream["r_frame_rate"]))

    # Compute total frame count safely
    if "nb_frames" in video_stream:
        total_frame_count = int(video_stream["nb_frames"])
    else:
        # Fallback: estimate using duration * FPS
        duration = float(probe["format"]["duration"])
        total_frame_count = int(duration * file_fps)

    if video_fps is not None:
        fps = video_fps
    else:
        fps = file_fps

    take_every = int(fps / reduce_fps_to) if reduce_fps_to and reduce_fps_to < fps else 1

    frame_size = width_height if width_height else original_resolution
    frame_bytes = frame_size.width * frame_size.height * 3  # RGB24

    # FFmpeg command setup
    input_kwargs = {}
    if starting_frame:
        input_kwargs["ss"] = starting_frame / fps  # Must be placed before "-i"

    output_kwargs = {
        "format": "rawvideo",
        "pix_fmt": "rgb24",
        "vsync": "0",  # Avoids unnecessary frame duplication
    }

    # Add scaling and FPS reduction to the filter chain
    filters: List[str] = []
    if width_height is not None:
        filters.append(f"scale={width_height.width}:{width_height.height}")

    if reduce_fps_to is not None:
        filters.append(f"fps={fps}")

    if filters:
        output_kwargs["vf"] = ",".join(filters)

    cmd = (
        ffmpeg.input(str(video_path), **input_kwargs)
        .output("pipe:", **output_kwargs)
        .global_args("-threads", "auto")  # Enables multi-threading
        .run_async(pipe_stdout=True, pipe_stderr=True, quiet=False)
    )

    def frames() -> Iterator[RGBInt8ImageType]:
        """
        Read frames off of the video capture until there none left or pulling a frame fails.
        :return: An iterator of frames.
        """

        try:
            while True:
                raw_frame = cmd.stdout.read(frame_bytes)
                if not raw_frame:
                    break
                frame = np.frombuffer(raw_frame, np.uint8).reshape(
                    (frame_size[1], frame_size[0], 3)
                )
                yield cast(RGBInt8ImageType, frame)
        finally:
            cmd.wait()

    return VideoFrames(
        original_fps=file_fps,
        original_resolution=original_resolution,
        total_frame_count=total_frame_count,
        frames=itertools.islice(frames(), None, None, take_every),
    )


# Starting place for hardware video decoder.
# pylint: disable=pointless-string-statement,trailing-whitespace
""" 
def frames_in_video_hardware(  # pylint: disable=too-many-locals
    video_path: Path,
    video_fps: Optional[float] = None,
    reduce_fps_to: Optional[float] = None,
    width_height: Optional[ImageResolution] = None,
    starting_frame: Optional[int] = None,
) -> VideoFrames:

    ffparams = {"-vcodec": "h264_cuvid", "-enforce_cv_patch": True}

    decoder = FFdecoder(
        str(video_path),
        frame_format="rgb24",
        verbose=True,
        **ffparams,
    ).formulate()

    metadata = json.loads(decoder.metadata)

    total_frame_count = metadata["approx_video_nframes"]
    file_fps = metadata["source_video_framerate"]

    if video_fps is not None:
        fps = video_fps
    else:
        fps = file_fps

    frame_count = itertools.count()

    if starting_frame is not None:
        frame_count = itertools.count(starting_frame)

    take_every = reduce_fps_take_every(original_fps=fps, new_fps=reduce_fps_to)

    original_width_height = ImageResolution(
        metadata["source_video_resolution"][0],
        metadata["source_video_resolution"][1],
    )

    resize = (original_width_height != width_height) if width_height is not None else False

    def frames() -> ImageSourceType:

        for frame in decoder.generateFrame():

            # check if frame is None
            if frame is None:
                break

            current_frame_index = next(frame_count)

            LOGGER.debug(
                f"Got a frame # {current_frame_index} / {total_frame_count} "
                f"({(current_frame_index / total_frame_count) * 100:.2f}) from: {video_path.name}"
            )

            # Convert YUV420p -> RGB
            rgb_frame = frame

            # Resize if needed
            if resize:
                rgb_frame = cv2.resize(rgb_frame, (width_height.width, width_height.height))

            yield cast(RGBInt8ImageType, rgb_frame)

    return VideoFrames(
        original_fps=fps,
        original_resolution=original_width_height,
        total_frame_count=total_frame_count,
        frames=itertools.islice(frames(), None, None, take_every),
    )
"""



"""
CLI-specific functionality.
"""

import random
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, Tuple, TypeVar

import click
from bonus_click.options import create_enum_option
from click import Context, Parameter
from click.decorators import FC
from click_option_group import MutuallyExclusiveOptionGroup, optgroup

from content_aware_timelapse.frames_to_vectors.conversion_types import (
    ConversionPOIsFunctions,
    ConversionScoringFunctions,
)
from content_aware_timelapse.frames_to_vectors.vector_computation.compute_vectors_clip import (
    CONVERT_CLIP,
)
from content_aware_timelapse.frames_to_vectors.vector_computation.compute_vectors_vit import (
    CONVERT_POIS_VIT_ATTENTION,
    CONVERT_SCORE_VIT_ATTENTION,
    CONVERT_SCORE_VIT_CLS,
)
from content_aware_timelapse.gpu_discovery import discover_gpus
from content_aware_timelapse.viderator.viderator_types import ImageResolutionParamType

# A generic type variable for Enum subclasses, essentially any enum subclass.
E = TypeVar("E", bound=Enum)
T = TypeVar("T")

ENV_VAR_PREFIX = "CAT"


def make_audio_option_group(name_prefix: str = "") -> Callable[[FC], FC]:
    """
    Defines the different methods to get set the audio that will be put under the resulting video.
    :param name_prefix: Prepended to the name the user will pass in for these options. Format is:
    `--{name_prefix}audio` / "--{name_prefix}audio-directory-random"
    :return: Function to become a click decorator that will produce the set of options.
    """

    def random_sort_audio_in_directory_callback(
        ctx: Context, _param: Parameter, value: Path
    ) -> None:
        """
        If the user provides a directory, collect all audio files within it,
        shuffle them randomly, and add the result to the "audio" option in the context.

        :param ctx: Click context
        :param _param: Click parameter (unused)
        :param value: Directory path provided via --random-audio-directory
        :return: None
        """

        if value is None:
            return None

        ctx.ensure_object(dict)

        # Gather supported audio files in the directory
        supported_exts = {".mp3", ".wav", ".flac", ".aiff", ".aac", ".ogg", ".m4a"}
        audio_files = [p for p in value.iterdir() if p.suffix.lower() in supported_exts]

        if not audio_files:
            raise click.BadParameter(f"No supported audio files found in directory: {value}")

        random.shuffle(audio_files)

        tuple_converted = tuple(audio_files)

        if ctx.params.get("audio", None) is None:
            ctx.params["audio"] = tuple_converted
        else:
            ctx.params["audio"] = ctx.params["audio"] + tuple_converted

        return None

    def add_paths_to_context(ctx: Context, _param: Parameter, value: Tuple[Path, ...]) -> None:
        """
        If the user provides a directory, collect all audio files within it,
        shuffle them randomly, and add the result to the "audio" option in the context.

        :param ctx: Click context
        :param _param: Click parameter (unused)
        :param value: Directory path provided via --random-audio-directory
        :return: None
        """

        ctx.ensure_object(dict)

        if ctx.params.get("audio", None) is None:
            ctx.params["audio"] = value
        else:
            ctx.params["audio"] = ctx.params["audio"] + value

    def decorator(command: FC) -> FC:
        """
        :param command: To wrap.
        :return: Decorated function.
        """

        group = optgroup.group(
            "Configures the audio that will be put under the resulting video",
            cls=MutuallyExclusiveOptionGroup,
        )

        for option in [
            optgroup.option(
                f"--{name_prefix}audio-directory-random",
                type=click.Path(
                    file_okay=False, exists=True, dir_okay=True, readable=True, path_type=Path
                ),
                help=(
                    "The audio files in this directory will be sorted in a "
                    "random order and added to resulting video."
                ),
                callback=random_sort_audio_in_directory_callback,
                envvar=f"{ENV_VAR_PREFIX}_AUDIO_DIRECTORY_RANDOM",
                expose_value=False,
                show_envvar=True,
            ),
            optgroup.option(
                f"--{name_prefix}audio",
                type=click.Path(
                    file_okay=True, exists=True, dir_okay=False, writable=True, path_type=Path
                ),
                help="If given, these audio(s) will be added to the resulting video.",
                multiple=True,
                envvar=f"{ENV_VAR_PREFIX}_AUDIO",
                callback=add_paths_to_context,
                expose_value=False,
                show_envvar=True,
            ),
        ]:
            option(command)

        group(command)

        return command

    return decorator


class VectorBackendScores(str, Enum):
    """
    For the CLI, string representations of the different vectorization backends.
    """

    vit_cls = "vit-cls"
    vit_attention = "vit-attention"
    clip = "clip"


class VectorBackendPOIs(str, Enum):
    """
    For the CLI, string representations of the different Point of Interest backends.
    """

    vit_attention = "vit-attention"


_CONVERSION_SCORING_FUNCTIONS_LOOKUP: Dict[VectorBackendScores, ConversionScoringFunctions] = {
    VectorBackendScores.vit_cls: CONVERT_SCORE_VIT_CLS,
    VectorBackendScores.vit_attention: CONVERT_SCORE_VIT_ATTENTION,
    VectorBackendScores.clip: CONVERT_CLIP,
}

_CONVERSION_POIS_FUNCTIONS_LOOKUP: Dict[VectorBackendPOIs, ConversionPOIsFunctions] = {
    VectorBackendPOIs.vit_attention: CONVERT_POIS_VIT_ATTENTION
}


input_files_arg = click.option(
    "--input",
    "-i",
    "input_files",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path),
    help="Input file(s). Can be given multiple times.",
    required=True,
    multiple=True,
)

output_path_arg = click.option(
    "--output-path",
    "-o",
    type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True, path_type=Path),
    help="Output will be written to this file.",
    required=True,
)

duration_arg = click.option(
    "--duration",
    "-d",
    type=click.FloatRange(min=1),
    help="Desired duration of the output video in seconds.",
    required=True,
    default=30.0,
    show_default=True,
)

output_fps_arg = click.option(
    "--output-fps",
    "-f",
    type=click.FloatRange(min=1),
    help="Desired frames/second of the output video.",
    required=True,
    default=60.0,
    show_default=True,
)

output_resolution_arg = click.option(
    "--output-resolution",
    "-or",
    type=ImageResolutionParamType(),
    help=(
        "Desired resolution of the output video. Video is resized as the final output step so "
        "resizing will occur after any cropping etc."
    ),
    required=False,
    default=None,
    show_default=True,
    envvar=f"{ENV_VAR_PREFIX}_OUTPUT_RESOLUTION",
    show_envvar=True,
)

frame_buffer_size_arg = click.option(
    "--frame-buffer-size",
    "-bu",
    type=click.IntRange(min=0),
    help=(
        "The number of frames to load into an in-memory buffer. "
        "This makes sure the GPUs have fast access to more frames rather than have the GPU "
        "waiting on disk/network IO."
    ),
    required=False,
    default=0,
    show_default=True,
    envvar=f"{ENV_VAR_PREFIX}_FRAME_BUFFER_SIZE",
    show_envvar=True,
)

viz_path_arg = click.option(
    "--viz-path",
    "-z",
    type=click.Path(file_okay=True, dir_okay=False, writable=True, path_type=Path),
    help=(
        "A visualisation describing the timelapse creation "
        "process will be written to this path if given"
    ),
    required=False,
)

deselect_arg = click.option(
    "--deselect",
    "-de",
    type=click.IntRange(min=0),
    help="Frames surrounding high scores will be dropped by a radius that starts with this value.",
    required=False,
    default=1000,
    show_default=True,
    envvar=f"{ENV_VAR_PREFIX}_DESELECT",
    show_envvar=True,
)

gpus_arg = click.option(
    "--gpu",
    "-g",
    type=click.Choice(choices=discover_gpus()),
    help="The GPU(s) to use for computation. Can be given multiple times.",
    required=False,
    multiple=True,
)

backend_pois_arg = create_enum_option(
    arg_flag="--backend-pois",
    help_message="Sets which Points of Interest discovery backend is used.",
    default=VectorBackendPOIs.vit_attention,
    input_enum=VectorBackendPOIs,
    lookup_fn=_CONVERSION_POIS_FUNCTIONS_LOOKUP.get,
    envvar=f"{ENV_VAR_PREFIX}_BACKEND_POIS",
)

backend_scores_arg = create_enum_option(
    arg_flag="--backend-scores",
    help_message="Sets which frame scoring backend is used.",
    default=VectorBackendScores.vit_cls,
    input_enum=VectorBackendScores,
    lookup_fn=_CONVERSION_SCORING_FUNCTIONS_LOOKUP.get,
    envvar=f"{ENV_VAR_PREFIX}_BACKEND_SCORES",
)

batch_size_pois_arg = click.option(
    "--batch-size-pois",
    "-bp",
    type=click.IntRange(min=1),
    help=(
        "Scaled frames for Points of Interest calculation are sent to GPU for "
        "processing in batches of this size."
    ),
    required=True,
    default=600,
    show_default=True,
    envvar=f"{ENV_VAR_PREFIX}_BATCH_POIS",
    show_envvar=True,
)

batch_size_scores_arg = click.option(
    "--batch-size-scores",
    "-bs",
    type=click.IntRange(min=1),
    help="Scaled frames for scoring are sent to GPU for processing in batches of this size.",
    required=True,
    default=100,  # Safe on many GPUs for both attention and scoring VIT.
    show_default=True,
    envvar=f"{ENV_VAR_PREFIX}_BATCH_SCORES",
    show_envvar=True,
)

vectors_path_pois_arg = click.option(
    "--vectors-path-pois",
    "-vp",
    type=click.Path(file_okay=True, dir_okay=False, writable=True, path_type=Path),
    help="Intermediate POI vectors will be written to this path. Can be used to re-run.",
    required=False,
)

vectors_path_scores_arg = click.option(
    "--vectors-path-scores",
    "-vs",
    type=click.Path(file_okay=True, dir_okay=False, writable=True, path_type=Path),
    help="Intermediate scoring vectors will be written to this path. Can be used to re-run.",
    required=False,
)

resize_inputs_arg = click.option(
    "--resize-inputs",
    "-ri",
    type=click.BOOL,
    help=(
        "If inputs are of different resolutions, the larger videos are shrunk to the smallest "
        "input resolution for analysis."
    ),
    required=False,
    default=True,
    show_default=True,
    envvar=f"{ENV_VAR_PREFIX}_RESIZE_INPUTS",
    show_envvar=True,
)


def video_inputs_outputs_args() -> Callable[[FC], FC]:
    """
    Common set of args supported by all the video manipulation functions.
    :return: Function to become a click decorator that will produce the set of options.
    """

    def decorator(command: FC) -> FC:
        """
        :param command: To wrap.
        :return: Decorated function.
        """

        for option in [
            input_files_arg,
            output_path_arg,
            duration_arg,
            output_fps_arg,
            output_resolution_arg,
            make_audio_option_group(),
            resize_inputs_arg,
        ]:
            option(command)

        return command

    return decorator
